'''
Parses class definitions out of GObject based headers

ToDo:
	- better parser aproach ;)
	- respect /*< public,protected,private >*/
	- parse *.c

'''
#    Copyright (c) 2006, Hans Breuer <hans@breuer.org>

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import glob, os, re

verbose = 0

def strip_whitespace (s) :
	s = s.replace (" ", "")
	return s.replace ("\t", "")

class CMethod :
	def __init__ (self, name, type) :
		self.name = name.strip()
		self.pars = []
		self.retval = type.replace(' ', '')
	def AddPar (self, name, type) :
		self.pars.append ((name.strip(),
					 strip_whitespace(type)))

class CClass :
	def __init__ (self, name) :
		self.name = name.strip()
		self.defs = {'TYPE' : None, 'OBJECT' : None, 'CLASS' : None}
		self.parent = None
		self.attrs = []
		self.methods = {}
		self.signals = []
	def AddAttr (self, name, type) :
		if verbose : print("\t", type, name)
		self.attrs.append ((name.strip(),
					  strip_whitespace(type)))
	def AddMethod (self, name, method) :
		if verbose : print("\t", name, "()")
		self.methods[name.strip()] = method
	def AddSignal (self, name, type) :
		self.signals.append ((name.strip(),
					    strip_whitespace(type)))

class CStripCommentsAndBlankLines :
	def __init__ (self, name) :
		f = open (name)
		lines = f.readlines ()
		self.lines = []
		self.cur = -1
		x1 = -1
		x2 = -1
		incom = 0
		r = re.compile("^\s*$") # to remove empty lines
		for s in lines :
			x1 = '/*'.find (s)
			x2 = '*/'.find (s)
			while x2 > x1 and incom != 1 :
				s = s[:x1] + s[x2+2:]
				x1 = '/*'.find (s)
				x2 = '*/'.find (s)
			else :
				if x1 > -1 :
					incom = 1
					s = s[:x1]
				elif x2 > - 1 and incom :
					s = s[x2+2:]
					incom = 0
			if r.match (s) or incom :
				continue
			self.lines.append (s)
		self.lines.append ("")
		f.close ()
	def readline (self) :
		self.cur = self.cur + 1
		try :
			return self.lines[self.cur]
		except :
			return ""
	def seek (self, toPos) :
		if toPos == 0 :
			self.cur = -1

def TestStripFile (name) :
	f = CStripCommentsAndBlankLines(name)
	s = f.readline()
	while s :
		print(s[:-1])
		s = f.readline()

import sys

sPkg = ''
sDir = '.'
if len (sys.argv) < 2 :
	print(sys.argv[0], '<package>', '[directory]')
	sys.exit(0)
else :
	sPkg = sys.argv[1]
if len (sys.argv) > 2 :
	sDir = sys.argv[2]

os.chdir (sDir)
lst  = glob.glob ('*.h')
lst.extend (glob.glob ("*/*.h"))
lst.extend (glob.glob ("*/*/*.h"))

klasses = []
no_klass_headers = []

#
# #define GDK_DRAWABLE(object) (G_TYPE_CHECK_INSTANCE_CAST ((object), GDK_TYPE_DRAWABLE, GdkDrawable))
#
rClassCast = re.compile ("^#define " + sPkg.upper () + "_(?P<n1>\w+)" \
				 "\((?P<par>\w+)\)\s+\(" \
				 "((G_TYPE_CHECK_INSTANCE_CAST)|(GTK_CHECK_CAST))\s+" \
				 "\(\((?P=par)\),\s*(?P<type>\w+),\s+" \
				 "(?P<name>\w+)\)\).*")
#
# m.group('type') == 'GDK_TYPE_DRAWABLE'
# m.group('name') == 'GdkTypeDrawable'
#
rVarDecl = re.compile ("\s*(?P<type>(unsigned)*\s*\w+\s*\**)\s*(?P<name>\w+)\s*:*\s*\d*\s*;\.*")
rSignal = rVarDecl #FIXME: signals are 'methods' to be registered
rMethod = re.compile ("\s*(?P<type>\w+\s+\**)\s*" \
			    "\(\*\s*(?P<name>\w+)\)\s*\(" \
			    "(?P<ptype>\w+\s*\**)\s*(?P<pname>\w+),*" \
			    "(?P<done>(\);)*)")
# m.group('type') == retval
# m.group('name')
# m.group('ptype'), m.group('pname')
# m.group('done')
rPar = re.compile ("\s*(const\s*)*(?P<ptype>\w+\s*\**)\s*(?P<pname>\w+),*" \
			 "(?P<done>(\);)*)")

fout = open (sPkg + '-generated.log', 'w')
# the type cast is not necessarily in the same header ;(
unresolved = {}
for sF in lst :
	sName = sF[len(sPkg):-2] # without case (approximation)
	# fin = open (sF)
	fin = CStripCommentsAndBlankLines(sF)
	s = fin.readline ()
	while s :
		m = rClassCast.match (s)
		if m :
			fout.write (m.group('type') + ' ' + m.group('name') + '\n')
			sName = m.group('name')
			unresolved[sName] = CClass (sName)
		s = fin.readline ()
	if len(list(unresolved.keys())) == 0 :
		no_klass_headers.append (sF)
	else :
		iFound = 0
		unresolved_keys = list(unresolved.keys())
		for sK in unresolved_keys :
			klass = unresolved[sK]
			sK = klass.name
			# start from the beginning
			fin.seek (0)
			rObject = re.compile ("struct\s+_" + sK + "\s*(?P<p>\{*)\s*$")
			rKlass  = re.compile ("struct\s+_" + sK + "Class\s*(?P<p>\{*)\s*$")
			s = fin.readline ()
			meth = None
			while s :
				mO = rObject.match (s)
				if mO :
					if mO.group('p') != '{' :
						s = fin.readline() # skip line with {
					s = fin.readline() # read parent line
					mV = rVarDecl.match(s)
					try :
						klass.parent = (mV.group('name'), mV.group('type').strip())
						if verbose : print("class", sK, ":", mV.group('type'))
					except :
						print('klass.parent error (', sF, ') :', s)
					s = fin.readline ()
					mV = rVarDecl.match(s)
					if not mV and verbose : print(s)
					while mV :
						klass.AddAttr (mV.group('name'), mV.group('type'))
						s = fin.readline ()
						mV = rVarDecl.match(s)

				else :
					mK = rKlass.match (s)
					if mK :
						iFound = iFound + 1
						klasses.append (unresolved[sK])
						del unresolved[sK]

						if mK.group('p') != '{' :
							s = fin.readline() # skip line with {
						s = fin.readline() # read parent line, validate it?
						s = fin.readline()
						mS = rSignal.match(s)
						mM = rMethod.match(s)
						while (mS or mM or meth) and s :
							if mS :
								klass.AddSignal (mS.group('name'), mS.group('type'))
							elif mM :
								meth = CMethod (mM.group('name'), mM.group('type'))
								klass.AddMethod (mM.group('name'), meth)
								meth.AddPar (mM.group('pname'), mM.group('ptype'))
								if mM.group('done') == ');' :
									meth = None # reset
							elif meth :
								mP = rPar.match (s)
								if mP :
									meth.AddPar (mP.group('pname'), mP.group('ptype'))
									if mP.group('done') == ');' :
										meth = None # reset
								else :
									# fixme: too drastic?
									#meth = None # reset
									pass
							else :
								break
							s = fin.readline ()
							mS = rSignal.match(s)
							mM = rMethod.match(s)
				s = fin.readline ()
		if iFound == 0 :
			no_klass_headers.append (sF)
		else :
			print(sF + " (" + str(iFound) + ")")

for sF in no_klass_headers :
	fout.write ("<No Klass>: " + sF + "\n")
print('Klasses found:', len(klasses), '\nHeaders w/o classes:', len(no_klass_headers))

def CmpParent (a,b) :
	'gets CClass, sort by parent type'
	if a.parent == None and b.parent == None :
		return 0
	elif a.parent == None :
		return 1
	elif b.parent == None :
		return -1
	try :
		i = cmp(a.parent[1],b.parent[1])
		if i == 0 :
			i = cmp(a.name, b.name)
		elif cmp(a.name, b.parent[1]) == 0 :
			#print a.name, ">", b.name
			return 1
		elif cmp(b.name, a.parent[1]) == 0 :
			#print b.name, ">", a.name
			return -1
		return i
	except :
		print("Sort Error:", str(a.parent), str(b.parent))
		return 0

klasses.sort(CmpParent)
# the sorting above does not sort everything, ensure parents are before childs
sorted = []
sorted_klasses = []
# first put in externals
parents = {}
for k in klasses :
	if k.parent :
		parents[k.parent[1]] = 1
for k in klasses :
	if k.name in parents :
		del parents[k.name]
sorted.extend(list(parents.keys()))
# sort the rest
while len(sorted_klasses) < len(klasses) :
	before = len(sorted_klasses)
	for k in klasses :
		sK = k.name
		sP = k.parent
		if sK in sorted :
			continue # don't add tem twice
		elif k.parent is None :
			sorted.append(sK)
		elif k.parent[1] in sorted :
			sorted.append(sK)
		else :
			continue
		#print sK
		sorted_klasses.append(k)
	if len(sorted_klasses) == before :
		if len(sorted_klasses) < len(klasses) :
			unsorted = []
			for k in klasses :
				if not k.name in sorted :
					unsorted.append(k.name)
			print(", ".join(unsorted), "not sorted?")
		break # avoid endless loop

klasses = sorted_klasses

def WritePython (fname) :
	# generate Python declaration for validation
	fpy = open (fname, "w")
	for klass in klasses :
		sParent = ""
		if klass.parent :
			sParent = " (" + klass.parent[1] + ")"
		fpy.write ('class ' + klass.name + sParent + " :\n")
		if len (klass.attrs) > 0 :
			fpy.write ('\tdef __init__ (self) :\n')
			for attr in klass.attrs :
				fpy.write ('\t\tself.' + attr[0] + " = None # " + attr[1] + "\n")
		if len (klass.signals) > 0 :
			fpy.write ('\t\t # Signals\n')
			for attr in klass.signals :
				fpy.write ('\t\tself.' + attr[0] + " = None # " + attr[1] + "\n")
		for s in list(klass.methods.keys()) :
			meth = klass.methods[s]
			fpy.write ('\t#returns: ' + meth.retval + '\n\tdef ' + meth.name + ' (')
			s1 = ''
			s2 = ''
			for par in meth.pars :
				s1 = s1 + par[0] + ', '
				s2 = s2 + par[1] + ', '
			if len(s1) > 0 :
				s1 = s1[:-2]
			if len(s2) > 0 :
				s2 = s2[:-2]
			fpy.write (s1 + ') :\n')
			fpy.write ('\t\t# ' + s2 + '\n')
			fpy.write ('\t\tpass\n')
		if len (klass.attrs) < 1 and len(klass.signals) < 1 and len(klass.methods) < 1 :
			fpy.write ('\tpass\n') # make it valid Python

def WriteDia (fname) :
	sStartDiagram = '''<?xml version="1.0" encoding="UTF-8"?>
<dia:diagram xmlns:dia="http://www.lysator.liu.se/~alla/dia/">
  <dia:layer name="Background" visible="true">'''
	sStartClass = '''
    <dia:object type="UML - Class" version="0" id="O%d">
      <dia:attribute name="elem_corner">
        <dia:point val="%d,%d"/>
      </dia:attribute>
      <dia:attribute name="name">
        <dia:string>#%s#</dia:string>
      </dia:attribute>
      <dia:attribute name="visible_attributes">
        <dia:boolean val="true"/>
      </dia:attribute>
      <dia:attribute name="visible_operations">
        <dia:boolean val="true"/>
      </dia:attribute>'''
	sFillColorAttribute = '''
      <dia:attribute name="fill_color">
        <dia:color val="%s"/>
      </dia:attribute>'''
	sStartAttributes = '''
      <dia:attribute name="attributes">
        <dia:composite type="umlattribute">'''
	sDataAttribute = '''
          <dia:attribute name="name">
            <dia:string>#%s#</dia:string>
          </dia:attribute>
          <dia:attribute name="type">
            <dia:string>#%s#</dia:string>
          </dia:attribute>'''
	sEndAttributes = '''
        </dia:composite>
      </dia:attribute>'''
	sStartOperations = '''
      <dia:attribute name="operations">'''
	sStartOperation = '''
        <dia:composite type="umloperation">
          <dia:attribute name="name">
            <dia:string>#%s#</dia:string>
          </dia:attribute>
          <dia:attribute name="type">
            <dia:string>#%s#</dia:string>
          </dia:attribute>
          <dia:attribute name="parameters">
            <dia:composite type="umlparameter">'''
	sDataParameter = '''
              <dia:attribute name="name">
                <dia:string>#%s#</dia:string>
              </dia:attribute>
              <dia:attribute name="type">
                <dia:string>#%s#</dia:string>
              </dia:attribute>'''
	sEndOperation = '''
            </dia:composite>
          </dia:attribute>
        </dia:composite>'''
	sEndOperations = '''
      </dia:attribute>'''
	sStartConnection = '''
    <dia:object type="UML - Generalization" version="1" id="O%d">
      <dia:attribute name="obj_pos">
        <dia:point val="%d,%d"/>
      </dia:attribute>
      <dia:attribute name="orth_autoroute">
        <dia:boolean val="true"/>
      </dia:attribute>'''
	sOrthPoints = '''
      <dia:attribute name="orth_points">
        <dia:point val="%d,%d"/>
        <dia:point val="%d,%d"/>
        <dia:point val="%d,%d"/>
        <dia:point val="%d,%d"/>
      </dia:attribute>'''
	sEndObject = '''
    </dia:object>'''
	sEndDiagram = '''
  </dia:layer>
</dia:diagram>'''

	fdia = open (fname, "w")
	nObject = 0
	x = 0
	y = 0
	dx = 10
	dy = 15
	# maintain a dictionary of parents positions to at least place not everything above each other
	positions = {}
	connectFrom = {}
	externals = {}
	fdia.write (sStartDiagram)

	for klass in klasses :
		if klass.parent : # add every parent ...
			parentName = klass.parent[1]
			if parentName in externals :
				externals[parentName] += 1
			else :
				externals[parentName] = 1
	for klass in klasses :
		if klass.name in externals : # ... but remove  the internals
			del externals[klass.name]

	# write all 'external' parents
	for s in list(externals.keys()) :
		externals[s] = (nObject, -1)
		fdia.write(sStartClass % (nObject, x, y, s))
		positions[s] = (x,y)
		x += dx
		fdia.write(sFillColorAttribute % ("#ffff00",))
		# fixme: any more attributes?
		fdia.write (sEndObject)
		nObject += 1

	for klass in klasses :
		parentName = ""
		if klass.parent :
			parentName = klass.parent[1]
			connectFrom[klass.name] = (nObject, parentName)
		if parentName in positions :
			x = positions[parentName][0] # same x
			y = positions[parentName][1] + dy # y below
		else :
			x += dx
			y = dy
		#fpy.write ('class ' + klass.name + sParent + " :\n")
		fdia.write(sStartClass % (nObject, x, y, klass.name))
		positions[klass.name] = (x, y)
		if len (klass.attrs) > 0 :
			fdia.write (sStartAttributes)
			for attr in klass.attrs :
				fdia.write (sDataAttribute % (attr[0], attr[1]))
			fdia.write (sEndAttributes)
#		if len (klass.signals) > 0 :
#			fpy.write ('\t\t # Signals\n')
#			for attr in klass.signals :
#				fpy.write ('\t\tself.' + attr[0] + " = None # " + attr[1] + "\n")
		# the differnence between signals and methods is in the attributes
		if len (klass.signals) > 0 or len(list(klass.methods.keys())) > 0  :
			fdia.write (sStartOperations)
		for s in list(klass.methods.keys()) :
			meth = klass.methods[s]
			fdia.write(sStartOperation % (meth.name, meth.retval))
			# first parameter is supposed to be 'this' pointer: leave out
			for par in meth.pars[1:] :
				fdia.write (sDataParameter % (par[0], par[1]))
			fdia.write (sEndOperation)
		if len (klass.signals) > 0 or len(list(klass.methods.keys())) > 0  :
			fdia.write (sEndOperations)
		fdia.write (sEndObject)
		nObject += 1

	# write all connections
	for sFrom in list(connectFrom.keys()) :
		iFrom = connectFrom[sFrom][0]
		sTo = connectFrom[sFrom][1]
		if sTo in connectFrom :
			iTo = connectFrom[sTo][0]
		elif sTo in externals :
			iTo = externals[sTo][0]
		else :
			print("sFrom -> sTo?", sFrom, sTo)
			continue # something wrong?
		nObject += 1
		#fdia.write ('\n\t<!-- %s : %s -->' % (sFrom, sTo))
		# just to give it some position (and stop Dia complaining)
		if sTo in positions :
			x1, y1 = positions[sTo]
		else :
			x1, y1 = (dx, dy)
		if sFrom in positions :
			x2, y2 = positions[sFrom]
		else :
			x2, y2 = (dx, dy)
		fdia.write (sStartConnection % (nObject, x1+dx/2, y1+dy/2,))
		fdia.write (sOrthPoints % (x1,y1, x1,(y1+y2)/2, x2,(y1+y2)/2, x2,y2 ))
		# and connect
		fdia.write ('''
      <dia:connections>
        <dia:connection handle="0" to="O%d" connection="6"/>
        <dia:connection handle="1" to="O%d" connection="1"/>
      </dia:connections>''' % (iTo, iFrom) )
		fdia.write (sEndObject)

	fdia.write (sEndDiagram)
	print(len(list(connectFrom.keys())), " connections")

WritePython (sPkg + "-generated.py")
WriteDia (sPkg + "-generated.dia")



#    PyDia Self Documentation Series - Part VII : All Sheets
#    Copyright (c) 2014  Hans Breuer  <hans@breuer.org>
#
#        generates something form the list of sheets
#

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
import dia, os
import tempfile
import webbrowser

import gettext
_ = gettext.gettext

# Given a list of "sheet objects" return the common namespace of the object types
def so_get_namespace (sol) :
	names = {}
	if len(sol) < 1 :
		return "Empty"
	for ot, descr, fname in sol :
		if ot :
			sp = ot.name.split(" - ")
			if len(sp) > 1 :
				if sp[0] in names:
					names[sp[0]] += 1
				else :
					names[sp[0]] = 1
	return ",".join (list(names.keys()))

def check_objecttype_overlap (sheets) :
	types = dia.registered_types()
	# remove Standard objects, they do not have or need a sheet
	del types["Group"]
	for s in ["Arc", "Box", "BezierLine", "Beziergon", "Ellipse", "Image", "Line",
		  "Outline", "Path", "Polygon", "PolyLine", "Text", "ZigZagLine"] :
		del types["Standard - %s" % (s,)]
	# got through all the sheets to match against registered types
	missing = []
	for sheet in sheets:
		for ot, descr, fname in sheet.objects:
			if ot.name in types:
				if ot == types[ot.name]:
					del types[ot.name]
				else:
					print("Mix-up:", ot.name)
			else:
				# sheet referencing a type not available
				missing.append(ot.name)
	# from the dictionary removed every type referenced just once?
	print(types)

def isheets_cb (data, flags) :
	sheets = dia.registered_sheets ()
	check_objecttype_overlap (sheets)
	path = tempfile.gettempdir() + os.path.sep + "dia-sheets.html"
	f = open (path, "w")
	f.write ("""
<html><head><title>Dia Sheets</title></head><body>
<table>
""")
	for sheet in sheets :
		info = "Namespace: [%s]<br>%i object types" % (so_get_namespace (sheet.objects), len(sheet.objects))
		sname = sheet.name
		if not sheet.user :
			sname = "<b>" + sname + "</b>"
		f.write ("<tr><td>%s</td><td>%s</td><td>%s</td></tr>\n" % (sname, sheet.description, info))
	f.write ("""</table>
</body></html>
""")
	dia.message(0, "'" + path + "' saved.")
	webbrowser.open('file://' + os.path.realpath(path))

dia.register_action ("HelpInspectSheets", _("Dia _Sheets Inspection"),
                     "/ToolboxMenu/Help/HelpExtensionStart",
                     isheets_cb)



# PyDia DOT Renderer
# Copyright (c) 2005 Hans Breuer <hans@breuer.org>

# A Dia export filter based on the object export interface.
# Generating output in the DOT language. That can be processed
# with the dot tool from http://www.graphviz.org

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, dia

import gettext
_ = gettext.gettext

class DotRenderer :
	def __init__ (self) :
		self.nodes = {}
		self.edges = {}
	def GetName (self, o) :
		"A small helper to turn a dia object into a name"
		s = ""
		if "name" in o.properties :
			s = o.properties["name"].value
		elif "text" in o.properties :
			s = o.properties["text"].value.text
		if s is None or s == "" :
			s = str(o)
		return s
	def GetColor (self, o) :
		if "fill_colour" in o.properties :
			rgb = o.properties["fill_colour"].value
			return "#%02x%02x%02x" % (rgb.red * 255, rgb.green * 255, rgb.blue * 255)
		return None
	def begin_render (self, data, filename) :
		self.f = open(filename, "w")
		self.f.write ("# %s\n" % (filename,))
		self.f.write ("# generated by (dia)dot.py\n")
		self.f.write ("digraph test {\n")
		# preserve the ratio as given by the bounding box
		rect = data.extents
		ratio = (rect.bottom - rect.top) / (rect.right - rect.left)
		self.f.write ("ratio=%.3f\n" % (ratio,))
		# extract some attributes as well as grouping
		self.f.write("node [style=filled]\n") # otherwise fillcolor is ignored
		for layer in data.layers :
			for o in layer.objects :
				if len(o.connections) < 1 :
					continue # no node
				if o.type.name == "Group" :
					continue # FIXME : iterate these, but we seem them below, too
				sColor = self.GetColor (o)
				if sColor and sColor != "#ffffff" : # dont write the default
					self.f.write('"%s" [fillcolor="%s"];\n' % (self.GetName(o), self.GetColor (o)))
		for layer in data.layers :
			for o in layer.objects :
				for c in o.connections :
					# these are the connection points of the objects. We will have e.g. "UML - Dependency"
					for n in c.connected :
						# we see the connecting lines multiple times, just process once
						if str(n) in self.edges :
							continue
						self.edges[str(n)] = 1
						if not (n.handles[0].connected_to and n.handles[1].connected_to) :
							continue
						# the right handles give us the direction
						a = n.handles[0].connected_to.object
						b = n.handles[1].connected_to.object
						try :
							self.f.write('"%s" -> "%s"\n' % (self.GetName(a), self.GetName(b)))
							#self.f.write('"%s" -> "%s"\n' % (str(a.properties["text"].value.text), str(b.properties["text"].value.text)))
						except :
							print(a, b, " writing connection failed.")
				else :
					pass
		self.f.write('}\n')
	def end_render (self) :
		self.f.close()

# dia-python keeps a reference to the renderer class and uses it on demand
dia.register_export (_("PyDia DOT Export"), "dot", DotRenderer())



#  PyDia SVG Renderer
#  Copyright (c) 2003, 2004 Hans Breuer <hans@breuer.org>
#

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, string, dia

import gettext
_ = gettext.gettext

##
# \brief The second SvgRenderer implementation for Dia
#
# A full blown SVG(Z) renderer. As of the initial writing less bugs in the output
# than the Dia SVG renderer written in C. Nowadays the _SvgRenderer is on par,
# but this one is still easier to extend and experiment with.
#
# \extends _DiaPyRenderer
# \ingroup ExportFilters
class SvgRenderer :
	def __init__ (self) :
		self.f = None
		self.line_width = 0.1
		self.line_caps = 0
		self.line_join = 0
		self.line_style = 0
		self.dash_length = 0
	def _open(self, filename) :
		self.f = open(filename, "w")
	def begin_render (self, data, filename) :
		self._open (filename)
		r = data.extents
		xofs = - r[0]
		yofs = - r[1]
		self.f.write('''<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!-- Created by diasvg.py -->
<svg width="%.3fcm" height="%.3fcm" viewBox="%.3f %.3f %.3f %.3f"
 xmlns="http://www.w3.org/2000/svg"
 xmlns:xlink="http://www.w3.org/1999/xlink">
''' % (r.right - r.left, r.bottom - r.top, r[0], r[1], r[2] - r[0], r[3] - r[1]))
		#self.f.write("<!-- %s -->\n" % (str(data.extents)))
		#self.f.write("<!-- %s -->\n" % (data.active_layer.name))
	def end_render (self) :
		self.f.write('</svg>')
		self.f.close()
	def is_capable_to (self, cap) :
		if cap == 1 : # RENDER_HOLES
			return 1
		elif cap == 2 : # RENDER_ALPHA
			return 1
		elif cap == 4 : #  RENDER_AFFINE
			return 1
		elif cap == 8 : # RENDER_PATTERN
			return 0
		return 0
	def draw_layer (self, layer, active, update) :
		self.f.write ("<!-- Layer: " + layer.name + " -->\n")
		self.f.write ('<g id="' + self._escape (layer.name) + '">\n')
		layer.render (self)
		self.f.write ('</g>\n')
	def draw_object (self, object, matrix) :
		self.f.write("<!-- " + object.type.name + " -->\n")
		odict = object.properties["meta"].value
		if "url" in odict :
			self.f.write('<a xlink:href="' + self._escape(odict["url"]) + '">\n')
		if "id" in odict or matrix :
			attrs = ''
			if matrix :
				attrs += 'transform="matrix' + str(matrix) + '" '
			if "id" in odict :
				attrs += 'id="' + self._escape(odict['id']) + '"'
			self.f.write('<g ' + attrs + '>\n')
		# don't forget to render the object
		object.draw (self)
		if "id" in odict or matrix :
			self.f.write('</g>\n')
		if "url" in odict :
			self.f.write('</a>\n')
	def set_linewidth (self, width) :
		if width < 0.001 : # zero line width is invisible ?
			self.line_width = 0.001
		else :
			self.line_width = width
	def set_linecaps (self, mode) :
		self.line_caps = mode
	def set_linejoin (self, mode) :
		self.line_join = mode
	def set_linestyle (self, style, length) :
		self.line_style = style
		self.dash_length = length
	def set_fillstyle (self, style) :
		# currently only 'solid' so not used anywhere else
		self.fill_style = style
	def set_font (self, font, size) :
		self.font = font
		self.font_size = size
	def draw_line (self, start, end, color) :
		self.f.write('<line x1="%.3f" y1="%.3f" x2="%.3f" y2="%.3f" stroke="%s" stroke-width="%.3f" %s/>\n' \
					% (start.x, start.y, end.x, end.y, self._rgb(color), self.line_width, self._stroke_style()))
	def draw_polyline (self, points, color) :
		self.f.write('<polyline fill="none" stroke="%s" stroke-width="%.3f" %s points="' \
					% (self._rgb(color), self.line_width, self._stroke_style()))
		for pt in points :
			self.f.write ('%.3f,%.3f ' % (pt.x, pt.y))
		self.f.write('"/>\n')
	def draw_polygon (self, points, fill, stroke) :
		self.f.write('<polygon fill="%s" stroke="%s" stroke-width="%.3f" %s points="' \
					% (self._rgb(fill), self._rgb(stroke), self.line_width, self._stroke_style()))
		for pt in points :
			self.f.write ('%.3f,%.3f ' % (pt.x, pt.y))
		self.f.write('"/>\n')
	def draw_rect (self, rect, fill, stroke) :
		self.f.write('<rect x="%.3f" y="%.3f" width="%.3f" height="%.3f" fill="%s" stroke="%s" stroke-width="%.3f" %s/>\n' \
					% (	rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top,
				  		self._rgb(fill), self._rgb(stroke), self.line_width, self._stroke_style()))
	def draw_rounded_rect (self, rect, fill, stroke, rounding) :
		self.f.write('<rect x="%.3f" y="%.3f" width="%.3f" height="%.3f" fill="%s" stroke="%s" stroke-width="%.3f" %s rx="%.3f" />\n' \
					% (	rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top,
				  		self._rgb(fill), self._rgb(stroke), self.line_width, self._stroke_style(), rounding))
	def _arc (self, center, width, height, angle1, angle2, color, fill=None) :
		# not in the renderer interface
		import math
		mPi180 = math.pi / 180.0
		rx = width / 2.0
		ry = height / 2.0
		sx = center.x + rx * math.cos(mPi180 * angle1)
		sy = center.y - ry * math.sin(mPi180 * angle1)
		ex = center.x + rx * math.cos(mPi180 * angle2)
		ey = center.y - ry * math.sin(mPi180 * angle2)
		largearc = abs(angle2 - angle1) >= 180
		sweep = 0 # 0: draw in negative direction
		if angle1 > angle2 :
			sweep = 1
		if not fill :
			self.f.write('<path stroke="%s" fill="none" stroke-width="%.3f" %s' \
				% (self._rgb(color), self.line_width, self._stroke_style()))
		else :
			self.f.write('<path stroke="none" fill="%s"' % (self._rgb(color)))
		# moveto sx,sy arc rx,ry x-axis-rotation large-arc-flag,sweep-flag ex,ey
		self.f.write(' d ="M %.3f,%.3f A %.3f,%.3f 0 %d,%d %.3f,%.3f ' % (sx, sy, rx, ry, largearc, sweep, ex, ey))
		self.f.write('"/>\n')
	def draw_arc (self, center, width, height, angle1, angle2, color) :
		self._arc(center, width, height, angle1, angle2, color)
	def fill_arc (self, center, width, height, angle1, angle2, color) :
		self._arc(center, width, height, angle1, angle2, color, 1)
	def draw_ellipse (self, center, width, height, fill, stroke) :
		self.f.write('<ellipse cx="%.3f" cy="%.3f" rx="%.3f" ry="%.3f" ' \
				'fill="%s" stroke="%s"  stroke-width="%.3f" %s/>\n' \
				% (center.x, center.y, width / 2, height / 2,
				   self._rgb(fill), self._rgb(stroke),
				   self.line_width, self._stroke_style()))
	def _bezier (self, bezpoints) :
		for bp in bezpoints :
			if bp.type == 0 : # BEZ_MOVE_TO
				self.f.write('M %.3f,%.3f ' % (bp.p1.x, bp.p1.y))
			elif bp.type == 1 : # BEZ_LINE_TO
				self.f.write('L %.3f,%.3f ' % (bp.p1.x, bp.p1.y))
			elif bp.type == 2 : # BEZ_CURVE_TO
				self.f.write ('C %.3f,%.3f %.3f,%.3f %.3f,%.3f ' \
					% (bp.p1.x, bp.p1.y, bp.p2.x, bp.p2.y, bp.p3.x, bp.p3.y))
			else :
				dia.message(2, "Invalid BezPoint type (%d)" * bp.type)
	def draw_bezier (self, bezpoints, color) :
		self.f.write('<path stroke="%s" fill="none" stroke-width="%.3f" %s d="' \
					% (self._rgb(color), self.line_width, self._stroke_style()))
		self._bezier (bezpoints)
		self.f.write('"/>\n')
	# old version, wont be called anymore with newer Dia
	def fill_bezier (self, bezpoints, color) :
		self.f.write('<path stroke="none" fill="%s" stroke-width="%.3f" d="' \
					% (self._rgb(color), self.line_width))
		self._bezier (bezpoints)
		self.f.write('"/>\n')
	def draw_beziergon (self, bezpoints, fill, stroke) :
		self.f.write('<path stroke="%s" fill="%s" stroke-width="%.3f" %s d="' \
				% (self._rgb(stroke), self._rgb(fill), self.line_width, self._stroke_style()))
		self._bezier (bezpoints)
		self.f.write('z"/>\n')
	def draw_string (self, text, pos, alignment, color) :
		if len(text) < 1 :
			return # shouldn'this be done at the higher level
		talign = ('start', 'middle', 'end') [alignment]
		fstyle = ('normal', 'italic', 'oblique') [self.font.style & 0x03]
		fweight = (400, 200, 300, 500, 600, 700, 800, 900) [(self.font.style  >> 4)  & 0x7]
		self.f.write('<text x="%.3f" y="%.3f"  fill="%s" text-anchor="%s" font-size="%.2f" font-family="%s" font-style="%s" font-weight="%d">\n' \
			% (pos.x, pos.y, self._rgb(color), talign, self.font_size, self.font.family, fstyle,  fweight))
		self.f.write(self._escape(text))
		self.f.write('</text>\n')
	def draw_image (self, point, width, height, image) :
		#FIXME : do something better than absolute pathes ?
		self.f.write('<image x="%.3f" y="%.3f"  width="%.3f" height="%.3f" xlink:href="%s"/>\n' \
			% (point.x, point.y, width, height, image.uri))
	# Helpers, not in the DiaRenderer interface
	def _escape (self, text) :
		# avoid writing XML special characters (ampersand must be first to not break the rest)
		for rep in [('&', '&amp;'), ('<', '&lt;'), ('>', '&gt;'), ('"', '&quot;'), ("'", '&apos;')] :
			text = text.replace (rep[0], rep[1])
		return text
	def _rgb(self, color) :
		# given a dia color convert to svg color string
		if not color :
			return "none"
		rgb = "#%02X%02X%02X" % (int(255 * color.red), int(color.green * 255), int(color.blue * 255))
		return rgb
	def _stroke_style(self) :
		# return the current line style as svg string
		dashlen =self.dash_length
		# dashlen/style interpretation like the DiaGdkRenderer
		dotlen = dashlen * 0.1
		caps = self.line_caps
		join = self.line_join
		style = self.line_style
		st = ""
		if style == 0 : # DIA_LINE_STYLE_SOLID
			pass
		elif style == 1 : # DASHED
			st = 'stroke-dasharray="%.2f,%.2f"' % (dashlen, dashlen)
		elif style == 2 : # DASH_DOT,
			gaplen = (dashlen - dotlen) / 2.0
			st = 'stroke-dasharray="%.2f,%.2f,%.2f,%.2f"' % (dashlen, gaplen, dotlen, gaplen)
		elif style == 3 : # DASH_DOT_DOT,
			gaplen = (dashlen - dotlen) / 3.0
			st = 'stroke-dasharray="%.2f,%.2f,%.2f,%.2f,%.2f,%.2f"' % (dashlen, gaplen, dotlen, gaplen, dotlen, gaplen)
		elif style == 4 : # DOTTED
			st = 'stroke-dasharray="%.2f,%.2f"' % (dotlen, dotlen)

		if join == 0 : # MITER
			pass # st = st + ' stroke-linejoin="bevel"'
		elif join == 1 : # ROUND
			st = st + ' stroke-linejoin="round"'
		elif join == 2 : # BEVEL
			st = st + ' stroke-linejoin="bevel"'

		if caps == 0 : # BUTT
			pass # default stroke-linecap="butt"
		elif caps == 1 : # ROUND
			st = st + ' stroke-linecap="round"'
		elif caps == 2 : # PROJECTING
			st = st + ' stroke-linecap="square"' # is this the same ?

		return st

class SvgzRenderer(SvgRenderer) :
	def _open(self, filename) :
		# There is some (here) not wanted behaviour in gzip.open/GzipFile :
		# the filename with path is not only used to adress the file but also
		# completely stored in the file itself. Correct it here.
		import os, os.path, gzip
		path, name = os.path.split(filename)
		os.chdir(path)
		self.f = gzip.open (name, "wb")

# dia-python keeps a reference to the renderer class and uses it on demand
dia.register_export (_("SVG plain"), "svg", SvgRenderer())
dia.register_export (_("SVG compressed"), "svgz", SvgzRenderer())



#  PyDia C++ Import
#  Copyright (c) 2006 Hans Breuer <hans@breuer.org>

# Another attempt to import C++ into a diagram. This time it is not trying to parse C++ directly
# but the XML generated by GCC_XML ( http://www.gccxml.org/ ).

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys

import gettext
_ = gettext.gettext

class Node :
	def __init__ (self, name) :
		self.name = name
		self.context = None
	def IsMethod (self) :
		return 0
	def IsUnion (self) :
		return 0
	def IsClass (self) :
		return 0
	def Name (self) :
		return self.name

class Type(Node) :
	def __init__ (self, name) :
		Node.__init__(self, name)
		self.names = []
	def AddName (self, name) :
		# for enumerations
		self.names.append(name)

class Union(Node) :
	def __init__ (self, name) :
		Node.__init__(self, name)
		self.names = []
	def IsUnion (self) :
		return 1
	def AddMember (self, name) :
		self.names.append(name)
	def Name (self) :
		ms = []
		# shortcut to avoid endless recursion with self referencing uions
		return self.name
		for s in self.names :
			ms.append (g_nodes[s].Name())
		return "; ".join(ms)

class Fassade(Node) :
	def __init__ (self, type, pre, post) :
		self.type = type
		self.pre = pre
		self.post = post
	def Name (self) :
		if self.pre != "" :
			return self.pre + " " + g_nodes[self.type].Name() + self.post
		else :
			return g_nodes[self.type].Name() + self.post

class Argument(Node) :
	def __init__ (self, name, type) :
		Node.__init__(self, name)
		self.type = type
		self.access = "public"
		self.static = 0
	def Type (self) :
		if self.type in g_nodes :
			return g_nodes[self.type].Name()
		return "?"
	def Visibility (self) :
		"This function is mapping from string to dia enum; should probably not be done here"
		if "protected" == self.access : return 2 #UML_PROTECTED
		elif "private" == self.access : return 1 #UML_PRIVATE
		return 0 #UML_PUBLIC

class Function(Node) :
	def __init__ (self, name, type) :
		Node.__init__(self, name)
		self.returns = type
		self.params = []
	def AddArg (self, arg) :
		self.params.append (arg)
	def Type (self) :
		if self.returns in g_nodes :
			return g_nodes[self.returns].Name()
		return ""
	def Signature (self) :
		args = []
		ret = ""
		for p in self.params :
			try :
				args.append (p.name + ":" + g_nodes[p.type].Name())
			except AttributeError :
				args.append (":" + p.Name())
			except :
				print("E:", p, p.name, p.type)
		if self.returns :
			ret = g_nodes[self.returns].Name() + " "
		return ret + self.name + " (" + ", ".join(args) + ")"

class Method(Function) :
	def __init__ (self, name, type) :
		Function.__init__(self, name, type)
		self.const = 0
		self.static = 0
		self.access = "public"
		self.virtual = 0
	def IsMethod (self) :
		return 1
	def Visibility (self) :
		"This function is mapping from string to dia enum; should probably not be done here"
		if "protected" == self.access : return 2 #UML_PROTECTED
		elif "private" == self.access : return 1 #UML_PRIVATE
		return 0 #UML_PUBLIC
	def InheritanceType (self) :
		if self.virtual > 1 : return 0 #UML_ABSTRACT
		if self.virtual > 0 : return 1 #UML_POLYMORPHIC
		return 2

class Klass(Node) :
	def __init__ (self, name) :
		Node.__init__(self, name)
		self.parents = []
		self.members = []
		self.abstract = 0
	def AddParent (self, id) :
		self.parents.append (id)
	def AddMember (self, id) :
		self.members.append (id)
	def IsClass (self) :
		return 1
	def Name (self) :
		if self.context in g_nodes and g_nodes[self.context].Name() != "" :
			return g_nodes[self.context].Name() + "::" + self.name
		else :
			return self.name
	def Parents (self) :
		# full qualified names
		names = []
		for p in self.parents :
			if p in g_nodes :
				names.append (g_nodes[p].Name())
		return names
	def Dump (self) :
		ps = ""
		for id in self.parents :
			if id in g_nodes :
				ps = ps + " " + g_nodes[id].Name()
		print(self.Name() + "(" + ps + " )")
		for id in self.members :
			print("\t", g_nodes[id], id)
			if g_nodes[id].IsMethod() :
				print("\t" + g_nodes[id].Signature())
			elif g_nodes[id].IsUnion() :
				print("\t" + g_nodes[id].Name())
			else :
				try :
					print("\t" + g_nodes[id].Name() + ":" + g_nodes[g_nodes[id].type].Name())
				except AttributeError :
					print("AttributeError:", g_nodes[id])

class Namespace(Node) :
	def __init__ (self, name) :
		Node.__init__(self, name)
		self.name = name
	def Name (self) :
		id = self.context
		if  id in g_nodes and g_nodes[id].Name() != "" :
			return g_nodes[id].Name() + "::" + self.name
		else :
			return self.name

g_nodes = {}
g_classes = []

def Parse (sFile, nodes) :

	import xml.parsers.expat
	global g_classes

	ctx = []
	def start_element(name, attrs) :
		o = None
		if name in ["Class", "Struct"] :
			#print attrs["name"], attrs["id"]
			o = Klass(attrs["name"])
			if "bases" in attrs :
				bs = " ".split (attrs["bases"])
				for s in bs :
					o.AddParent (s)
			if "members" in attrs :
				ms = " ".split (attrs["members"])
				for s in ms :
					if s != "" :
						o.AddMember (s)
			if "abstract" in attrs :
				o.abstract = int(attrs["abstract"])
			g_classes.append (o)
		elif "Union" == name :
			o = Union(attrs["name"])
			if "members" in attrs :
				ms = attrs["members"].split (" ")
				for s in ms :
					if s != "" :
						o.AddMember (s)
			# FIXME: this creates a dup
		elif "Namespace" == name :
			#print attrs["name"], attrs["id"], "::"
			if attrs["name"] == "::" :
				o = Namespace("")
			else :
				o = Namespace(attrs["name"])
		elif name in ["Method", "OperatorMethod", "Constructor", "Destructor"] :
			if "Constructor" == name : o = Method (attrs["name"], None)
			elif "Destructor" == name : o = Method ("~" + attrs["name"], None)
			else : o = Method (attrs["name"], attrs["returns"])

			if "virtual" in attrs : o.virtual += int(attrs["virtual"])
			if "pure_virtual" in attrs : o.virtual += int(attrs["pure_virtual"])
			if "access" in attrs : o.access = attrs["access"]
		elif name in ["Field", "Typedef"] :
			o = Argument (attrs["name"], attrs["type"])
			if "access" in attrs : o.access = attrs["access"]
			if "static" in attrs : o.static = attrs["static"]
		elif name in ["FundamentalType", "Enumeration"] :
			o = Type (attrs["name"])
		elif "ReferenceType" == name :
			o = Fassade (attrs["type"], "", "&")
		elif "PointerType" == name :
			o = Fassade (attrs["type"], "", "*")
		elif "ArrayType" == name :
			o = Fassade (attrs["type"], "", "[]")
		elif "CvQualifiedType" == name :
			o = Fassade (attrs["type"], "const", "")
		elif "Argument" == name :
			if "name" in attrs :
				o = Argument (attrs["name"], attrs["type"])
			else :
				o = Fassade (attrs["type"], "", "")
			if ctx[-1][1] :
				ctx[-1][1].AddArg (o)
			o = None # lookup not possible
		elif "Ellipsis" == name :
			o = Argument ("", "...")
			if ctx[-1][1] :
				ctx[-1][1].AddArg (o)
			o = None # lookup not possible
		elif "EnumValue" == name :
			if ctx[-1][1] :
				ctx[-1][1].AddName (attrs["name"])
		elif name in ["Function", "OperatorFunction", "FunctionType"] :
			if "name" in attrs :
				o = Function (attrs["returns"], attrs["name"])
			else : # function & type
				o = Function (attrs["returns"], attrs["id"])
		elif "Variable" == name :
			pass # not ofinterest?
		elif "File" == name :
			pass # FIXME: thrown away
		else :
			print("Unhandled:", name)
		if o :
			if "context" in attrs :
				#print attrs["context"]
				o.context = attrs["context"]
			nodes[attrs["id"]] = o
		ctx.append((name, o)) #push

	def end_element(name) :
		del ctx[-1] # pop
	def char_data(data) :
		pass

	p = xml.parsers.expat.ParserCreate()
	p.StartElementHandler = start_element
	p.EndElementHandler = end_element
	p.CharacterDataHandler = char_data

	p.Parse(sFile)

def Process (sName) :
	global g_nodes
	f = open (sName)
	Parse(f.read(), g_nodes)

def ImportCpp (sFile, diagramData) :
	# process the c++ file with GCC_XML to get XML
	# TODO
	ImportXml (sXmlFile, diagramData)

def ImportXml (sFile, diagramData) :
	# XML to internal representation
	global g_nodes, g_classes
	g_nodes = {} # these are not interchangeable between diagrams
	g_classes = [] # neither are these
	Process (sFile)
	# internal representation to diagram
	layer = diagramData.active_layer
	# we need some kind of filter to not always generate the whole stl diagram
	theLinks = {}
	nTotal = 0
	for c in g_classes :
		theLinks[c.Name()] = 0
	for c in g_classes :
		bUsed = 0
		for p in c.Parents() :
			if c.Name()[:5] == "std::" : continue # is this too drastic ?
			if p in theLinks :
				theLinks[p] += 1
				bUsed = 1
		if bUsed :
			# to have bottom most visible as well
			theLinks[c.Name()] += 1
			nTotal += 1
	if nTotal < 2 : # arbitrary limit to generate simple diagrams not using inheritance at all
		for c in g_classes :
			if c.Name in theLinks :
				theLinks[c.Name()] += 1
	# now everything interesting should be in theLinks with a 'ref count' above zero
	for c in g_classes :
		if c.Name() not in theLinks : continue
		if theLinks[c.Name()] :
			theLinks[c.Name()] = c
		else :
			del theLinks[c.Name()]
	theObjects = {}
	for s in list(theLinks.keys()) :
		o, h1, h2 = dia.get_object_type("UML - Class").create(0,0)
		layer.add_object(o)
		o.properties["name"] = s.encode("UTF-8")
		if c.abstract :
			o.properties["abstract"] = 1
		methods = []
		attributes = []
		c = theLinks[s]
		for mid in c.members :
			if mid not in g_nodes :
				continue #HACK
			m = g_nodes[mid]
			#print m
			if m.IsMethod () : # (name, type, comment, stereotype, visibility, inheritance_type, query,class_scope, params)
				params = []
				for a in m.params :
					# (name, type, value, comment, kind)
					try  :
						print(a.name, a.Type())
						params.append ((a.name.encode("UTF-8"), a.Type().encode("UTF-8"), None, "", 0))
					except :
						pass
				methods.append ((m.name.encode("UTF-8"), m.Type().encode("UTF-8"), "", "", m.Visibility(),m.InheritanceType(),0,0, params))
			elif m.IsUnion () :
				pass
			else : # (name,type,value,comment,visibility,abstract,class_scope)
				try  :
					attributes.append ((m.Name().encode("UTF-8"), m.Type().encode("UTF-8"), "", "", m.Visibility(),0,m.static))
				except  :
					print("Error", m.name)
		# set some properties
		o.properties["operations"] = methods
		o.properties["attributes"] = attributes

		theObjects[s] = o
	# class connections
	for s in list(theLinks.keys()) :
		o1 = theObjects[s]
		c = theLinks[s]
		for p in c.Parents() :
			o, h1, h2 = dia.get_object_type("UML - Generalization").create(0,0)
			layer.add_object(o)
			o2 = theObjects[p]
			h1.connect (o2.connections[6])
			h2.connect (o1.connections[1])
	# update placement depending on number of parents ?

	layer.update_extents()
	#dia.active_display().add_update_all()

if __name__ == '__main__':
	Process(sys.argv[1])
	for c in g_classes :
		if c.Name()[:2] != "__" and c.Name()[:5] != "std::" :
			try :
				if c.IsClass() :
					c.Dump()
			except KeyError :
				pass
else :
	try :
		import dia
		#dia.register_import("Cpp via GCC_XML", "cpp", ImportCpp)
		dia.register_import(_("XML from GCC_XML"), "xml", ImportXml)
	except ImportError :
		pass



#    PyDia Self Documentation Series - Part IV : All Objects Properties
#    Copyright (c) 2008, Hans Breuer <hans@breuer.org>
#
#        generates a new diagram which contains all the currently
#    known std-props, that is iterates all objects for their properties
#

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, dia
from functools import cmp_to_key

import gettext
_ = gettext.gettext

def allprops_cb(data, flags) :

	# copied from otypes.py
	if data:
		diagram = None # we may be running w/o GUI
	else :
		diagram = dia.new("All Object Properties.dia")
		data = diagram
	layer = data.active_layer

	props_by_name = {}
	otypes = dia.registered_types()
	name_type_clashes = []

	for oname in otypes :
		try :
			obj, h1, h2 = dia.get_object_type(oname).create (0, 0)
		except :
			print("Huh?", oname)
			continue
		prop_keys = list(obj.properties.keys())
		for k in prop_keys :
			p = obj.properties[k]
			if k in props_by_name :
				# check if it is the same type
				p0, names = props_by_name[k]
				try :
					if p0.type != p.type :
						# construct a unique name
						uname = p.name + "<" + p.type + ">"
						if uname in props_by_name :
							props_by_name[uname][1].append(oname)
						else :
							props_by_name[uname] = (p, [oname])
							name_type_clashes.append (oname + "::" + p.name + " as " + p0.type + " and " + p.type)
					else :
						# remember the origin of the property
						props_by_name[k][1].append(oname)
				except KeyError :
					print(oname, "::", k, p, "?")
			else :
				props_by_name[k] = (p, [oname])
		obj.destroy() # unsave delete, any method call will fail/crash afterweards
		obj = None

	grid = {} # one type per row
	dx = 1.0
	dy = 5.0
	ot = dia.get_object_type("UML - Class")

	props_keys = list(props_by_name.keys())
	# alpha-numeric sorting by type; after by number of users
	props_keys.sort (key=cmp_to_key (lambda a,b : len(props_by_name[b][1]) - len(props_by_name[a][1])))
	props_keys.sort (key=lambda a: props_by_name[a][0].type)

	almost_all = 98 * len(otypes) / 100 # 98 %

	for pname in props_keys :

		p, names = props_by_name[pname]

		x = 0.0
		y = 0.0
		if p.type in grid:
			x, y = grid[p.type]
		else :
			x = 0.0
			y = len(list(grid.keys())) * dy
		o, h1, h2 = ot.create (x,y)
		o.properties["name"] = pname
		o.properties["template"] = 1
		o.properties["templates"] = [(p.type, '')]
		# coloring depending on use
		if len(names) > almost_all :
			o.properties["fill_colour"] = "lightgreen"
		elif len(names) > 100 :
			o.properties["fill_colour"] = "lightblue"
		elif len(names) > 2 :
			o.properties["fill_colour"] = "lightcyan"
		elif len(names) > 1 :
			o.properties["fill_colour"] = "lightyellow"
		# if there is only one user show it
		if len(names) == 1 :
			o.properties["comment"] = names[0]
			o.properties["visible_comments"] = 1
			o.properties["comment_line_length"] = 60
		else :
			o.properties["comment"] = "; ".join(names)
			o.properties["visible_comments"] = 0
			o.properties["comment_line_length"] = 60

		# store position for next in row
		x += (dx + o.properties["elem_width"].value)
		grid[p.type] = (x,y)

		layer.add_object(o)
	layer.update_extents()
	data.update_extents()
	if diagram :
		diagram.display()
		diagram.flush()
	if len(name_type_clashes) > 0 :
		dia.message(0, "One name, one type?!\n" + "\n".join(name_type_clashes))
	return data

dia.register_action ("HelpAllPropts", _("All Object _Properties"),
                     "/ToolboxMenu/Help/HelpExtensionStart",
                     allprops_cb)



#
# Dissect a diagram by rendering it and accumulating invalid
# renderer calls to object selection.
#
# Copyright (c) 2014 Hans Breuer <hans@breuer.org>
#
#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, string, dia

import gettext
_ = gettext.gettext

##
# \brief A dissecting renderer for Dia
#
# Check the diagram by rendering it and report anomalies with their
# call and the causing object.
#
# \extends _DiaPyRenderer
# \ingroup ExportFilters
class DissectRenderer :
	def __init__ (self) :
		self.f = None
		self.current_objects = []
		self.warnings = []
		self.errors = []
		self.font = None
	def _open(self, filename) :
		self.f = open(filename, "w")
	def begin_render (self, data, filename) :
		self._open (filename)
		self.extents = data.extents
		try :
			# this can fail when data is not interative,
			# e.g. when running from pure bindings
			self.f.write ("# Dissect %s\n" % (data.filename,))
		except :
			self.f.write ("# Dissect %s\n" % (filename,))
	def end_render (self) :
		self.f.write('%d error(s) %d warning(s)\n' % (len(self.errors), len(self.warnings)))
		self.f.close ()
	def Warning (self, msg) :
		self.warnings.append ((self.current_objects[-1], msg))
		if self.f :
			self.f.write ("Warning: %s, %s\n" % (self.current_objects[-1], msg))
	def Error (self, msg) :
		self.errors.append ((self.current_objects[-1], msg))
		if self.f :
			self.f.write ("Error: %s, %s\n" % (self.current_objects[-1], msg))
	def draw_object (self, object, matrix) :
		self.current_objects.append (object)
		# XXX: check matrix
		# don't forget to render the object
		object.draw (self)
		del self.current_objects[-1]
	def set_linewidth (self, width) :
		# width==0 is hairline
		if width < 0 or width > 10 :
			self.Warning ("linewidth out of range")
	def set_linecaps (self, mode) :
		if mode < 0 or mode > 2 :
			self.Error ("linecaps '%d' unknown" % (mode,))
	def set_linejoin (self, mode) :
		if mode < 0 or mode > 2 :
			self.Error ("linejoin '%d' unknown" % (mode,))
	def set_linestyle (self, style, dash_length) :
		if style < 0 or style > 4 :
			self.Error ("linestyle '%d' unknown" % (style,))
		if dash_length < 0.001 or dash_length > 1 :
			self.Warning ("dashlength '%f' out of range" % (dash_length,))
	def set_fillstyle (self, style) :
		# currently only 'solid' so not used anywhere else
		if style != 0 :
			self.Error ("fillstyle '%d' unknown" % (style,))
	def set_font (self, font, size) :
		self.font = font
		self.font_size = size
	def draw_line (self, start, end, color) :
		pass # can anything go wrong here ?
	def draw_polyline (self, points, color) :
		if len(points) < 2 :
			self.Error ("draw_polyline with too few points")
	def _polygon (self, points, fun) :
		if len(points) < 3 :
			self.Error ("%s with too few points" % (fun,))
	def draw_polygon (self, points, fill, stroke) :
		self._polygon(points, "draw_polygon")
	# obsolete with recent Dia
	def fill_polygon (self, points, color) :
		self._polygon(points, "draw_polygon")
	def _rect (self, rect, fun) :
		if rect.top > rect.bottom :
			self.Warning ("%s negative height" % (fun,))
		if rect.left > rect.right :
			self.Warning ("%s negative width" % (fun,))
	def draw_rect (self, rect, fill, stroke) :
		self._rect (rect, "draw_rect")
	def draw_rounded_rect (self, rect, fill, stroke, rounding) :
		# XXX: check rounding to be positive (smaller than half width, height?)
		self._rect (rect, "draw_rect")
	def _arc (self, center, width, height, angle1, angle2, fun) :
		if width <= 0 :
			self.Warning ("%s width too small" % (fun,))
		if height <= 0 :
			self.Warning ("%s height too small" % (fun,))
		# angles
		rot = 0.0
		if angle1 < angle2 :
			rot = angle2 - angle1
		else :
			rot = angle1 - angle2
		if rot <= 0 or rot >= 360 :
			self.Warning ("%s bad rotation %g,%g" % (fun, angle1, angle2))
	def draw_arc (self, center, width, height, angle1, angle2, color) :
		self._arc(center, width, height, angle1, angle2, "draw_arc")
	def fill_arc (self, center, width, height, angle1, angle2, color) :
		self._arc(center, width, height, angle1, angle2, "fill_arc")
	def draw_ellipse (self, center, width, height, fill, stroke) :
		self._arc(center, width, height, 0, 360, "draw_ellipse")
	def _bezier (self, bezpoints, fun) :
		nMoves = 0
		for bp in bezpoints :
			if bp.type == 0 : # BEZ_MOVE_TO
				nMoves = nMoves + 1
				if nMoves > 1 :
					self.Warning ("%s move-to within", (fun,))
			elif bp.type == 1 : # BEZ_LINE_TO
				pass
			elif bp.type == 2 : # BEZ_CURVE_TO
				pass
			else :
				self.Error ("%s invalid BezPoint type='%d'" % (fun, bp.type,))
	def draw_bezier (self, bezpoints, color) :
		if len(bezpoints) < 2 :
			self.Error ("draw_bezier too few points");
		self._bezier (bezpoints, "draw_bezier")
	def fill_bezier (self, bezpoints, color) :
		if len(bezpoints) < 3 :
			self.Error ("fill_bezier too few points");
		self._bezier (bezpoints, "fill_bezier")
	def draw_string (self, text, pos, alignment, color) :
		if len(text) < 1 :
			self.Warning ("draw_string empty text")
		if alignment < 0 or alignment > 2 :
			self.Error ("draw_string unknown alignmern '%d'" % (alignment,))
	def draw_image (self, point, width, height, image) :
		if width <= 0 :
			self.Warning ("draw_image width too small")
		if height <= 0 :
			self.Warning ("draw_image height too small")
		# XXX: check image, e.g. existing file name
# dia-python keeps a reference to the renderer class and uses it on demand
dia.register_export (_("Dissect"), "dissect", DissectRenderer())



#   Un-inline (i.e. save) embedded images
#
#   Copyright (c) 2011, Hans Breuer <hans@breuer.org>
#
#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os, string, sys, dia

import gettext
_ = gettext.gettext

class UninlineRenderer :
	def __init__ (self) :
		self.count = 0
	def begin_render (self, data, filename) :
		imgmap = {}
		dirname = os.path.dirname (filename)
		basename = os.path.basename(filename)
		ext = filename[filename.rfind(".")+1:]
		for layer in data.layers :
			for o in layer.objects :
				if "inline_data" in list(o.properties.keys()) :
					if o.properties["inline_data"].value :
						# remember by position
						pos = o.properties["obj_pos"].value
						xk = "%03g" % pos.x
						yk = "%03g" % pos.y
						key = basename + "-" + layer.name + "x" + xk + "y" + yk
						imgmap[key] = o
			for k, o in imgmap.items() :
				fname = dirname + "/" + k + "." + ext
				print(fname)
				o.properties["image_file"] = fname
				o.properties["inline_data"] = 0

	def end_render (self) :
		pass

# dia-python keeps a reference to the renderer class and uses it on demand
dia.register_export (_("Uninline Images"), "png", UninlineRenderer())



import sys, dia

import gettext
_ = gettext.gettext

# sys.path.insert(0, 'd:/graph/dia/dia')

class ObjRenderer :
	def __init__ (self) :
		pass
	def begin_render (self, data, filename) :
		self.f = open(filename, "w")
		for layer in data.layers :
			self.f.write ("<layer name=\"" + layer.name + "\">\n")
			for o in layer.objects :
				self.f.write ("\t"*1 + str(o) + "\n")
				r = o.bounding_box
				self.f.write ("\t"*2 + "<bbox>" + str(r.left) + "," + str(r.top) + ";" \
						+ str(r.right) + "," + str(r.bottom) + "</bbox>\n")

				self.f.write ("\t"*2 + "<connections>\n")
				for c in o.connections :
					if (len(c.connected) < 1) :
						continue
					self.f.write ("\t"*3 + str(c) + "\n")
					for co in c.connected :
						self.f.write ("\t"*4 + str(co) + "\n")
					self.f.write ("\t"*3 + "</DiaConnectionPoint>\n")
				self.f.write ("\t"*2 + "</connections>\n")

				self.f.write ("\t"*2 + str(o.properties) + "\n")
				keys = list(o.properties.keys())
				for s in keys :
					self.f.write ("\t"*3 + str(o.properties[s]) + "\n")
					if o.properties[s].type == "string" :
						self.f.write ("\t"*3 + str(o.properties[s].value) + "\n")
					self.f.write ("\t"*3 + "</DiaProperty>\n")
				self.f.write ("\t"*2 + "</DiaProperties>\n")
				self.f.write ("\t"*1 + "</DiaObject>\n\n")
			self.f.write ("\t"*0 + "</layer>\n")
	def end_render (self) :
		self.f.close()

# dia-python keeps a reference to the renderer class and uses it on demand
dia.register_export (_("PyDia Object Export"), "diapyo", ObjRenderer())



# PyDia Rotation
# Copyright (c) 2003, Hans Breuer <hans@breuer.org>
# Copyright (c) 2009, 2011  Steffen Macke <sdteffen@sdteffen.de
#
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import warnings
import dia, math

import gettext
_ = gettext.gettext

class CRotateDialog :
	def __init__(self, data) :
		import gi

		gi.require_version('Gtk', '3.0')

		with warnings.catch_warnings():
			warnings.filterwarnings("ignore", category=RuntimeWarning)
			from gi.repository import Gtk

		win = Gtk.Window()
		win.connect("delete_event", self.on_delete)
		win.set_title(_("Rotate counter-clockwise"))

		self.diagram = data
		self.win = win

		box1 = Gtk.VBox()
		win.add(box1)
		box1.show()

		box2 = Gtk.VBox(spacing=10)
		box2.set_border_width(10)
		box1.pack_start(box2, True, True, 0)
		box2.show()

		label1 = Gtk.Label()
		label1.set_text(_('Rotation around (0,0). Rotation angle in degrees:'))
		box2.pack_start(label1, True, True, 0)
		label1.show()

		self.entry = Gtk.Entry()
		self.entry.set_text("0.0")
		box2.pack_start(self.entry, True, True, 0)
		self.entry.show()

		separator = Gtk.HSeparator()
		box1.pack_start(separator, 0, True, 0)
		separator.show()

		box2 = Gtk.VBox(spacing=10)
		box2.set_border_width(10)
		box1.pack_start(box2, 0, True, 0)
		box2.show()

		button = Gtk.Button("rotate")
		button.connect("clicked", self.on_rotate)
		box2.pack_start(button, True, True, 0)
		button.set_can_default(True)
		button.grab_default()
		button.show()
		win.show()

	def on_rotate(self, *args) :
		s = self.entry.get_text()
		angle = float(s)
		if angle >= 0 and angle <= 360 :
			SimpleRotate (self.diagram, angle)
			self.diagram.update_extents()
			self.diagram.flush()
		else :
			dia.message(1, "Please enter an angle between 0 and 360 degrees.")
		self.win.destroy ()

	def on_delete (self, *args) :
		self.win.destroy ()

def SimpleRotate(data, angle) :
	# Rotation center
	xm = 0.0
	ym = 0.0

	# Convert to radians
	angle_rad = 2*math.pi - 2*math.pi*angle/360

	objs = data.get_sorted_selected()
	if len(objs) == 0 :
		objs = data.active_layer.objects
	scaleFailed = {}
	ptype = dia.get_object_type('Standard - Polygon')
	for o in objs :
		if o.type.name == 'Standard - Box' :
			r = o.properties['obj_bb'].value
			p = ptype.create(0,0)
			p = p[0]
			p.properties['poly_points'] = [(r.left, r.top), (r.right, r.top), (r.right, r.bottom), (r.left, r.bottom)]
			p.properties['line_width'] = o.properties['line_width']
			p.properties['line_colour'] = o.properties['line_colour']
			p.properties['line_style'] = o.properties['line_style']
			p.properties['line_colour'] = o.properties['line_colour']
			p.properties['fill_colour'] = o.properties['fill_colour']
			p.properties['show_background'] = o.properties['show_background']
			data.active_layer.add_object(p)
			data.active_layer.remove_object(o)
			o = p
		for h in o.handles:
			x = math.cos(angle_rad)*(h.pos.x+xm)-math.sin(angle_rad)*(h.pos.y+ym)
			y = math.sin(angle_rad)*(h.pos.x+xm)+math.cos(angle_rad)*(h.pos.y)
			o.move_handle(h, (x,y), 0, 0)

	data.update_extents ()
	dia.active_display().add_update_all()

def rotate_cb(data, flags) :
	dlg = CRotateDialog(data)

dia.register_action ("ObjectsSimplerotation", _("Simple _Rotation"),
		     "/DisplayMenu/Objects/ObjectsExtensionStart",
		     rotate_cb)



#  PyDia Standard Dia Renderer
#  Copyright (c) 2009 Hans Breuer <hans@breuer.org>
#
#  A renderer which outputs Dia "Standard - *" Objects in Dia XML
#  by implementing the renderer interface.

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, string, os, math, dia

import gettext
_ = gettext.gettext

class StandardDiaRenderer :
	def __init__ (self) :
		self.f = None
		self.line_width = 0.1
		self.line_caps = 0
		self.line_join = 0
		self.line_style = 0
		self.dash_length = 0
		# every object should have a unique id - mainly for connections definition
		self.oid = 0
		self.filename = ""
	def _open(self, filename) :
		self.f = open(filename, "w")
	def begin_render (self, data, filename) :
		self.filename = filename
		self._open (filename)
		portrait = "false"
		if data.paper.is_portrait :
			portrait = "true"
		# data.paper doesn't provide margins yet, data.paper.width is _useable_ width
		self.f.write('''<?xml version="1.0" encoding="UTF-8"?>
<dia:diagram xmlns:dia="http://www.lysator.liu.se/~alla/dia/">
  <!-- Created by dia_standard_dia.py -->
  <dia:diagramdata>
    <!-- everything is optional here -->
    <dia:attribute name="paper">
      <dia:composite type="paper">
        <dia:attribute name="name">
          <dia:string>#%s#</dia:string>
        </dia:attribute>
        <dia:attribute name="is_portrait">
          <dia:boolean val="%s"/>
        </dia:attribute>
        <dia:attribute name="scaling">
          <dia:real val="%.3f"/>
        </dia:attribute>
      </dia:composite>
    </dia:attribute>
  </dia:diagramdata>
  <dia:layer name="Background" visible="true" >
''' % (data.paper.name, portrait, data.paper.scaling))
	def end_render (self) :
		self.f.write('''<!-- no connections in the renderer interface -->
  </dia:layer>
</dia:diagram>''')
		self.f.close()
	# some helpers
	def _rgb(self, color) :
		# given a dia color convert to svg color string
		rgb = "#%02X%02X%02X" % (int(255 * color.red), int(color.green * 255), int(color.blue * 255))
		return rgb
	# http://bugzilla.gnome.org/show_bug.cgi?id=576468 makes this even less desirable
	# see also: samples/Self/dia-standard-objects.dia for inconsistency mapping
	def _tinting (self, color, fill=None, stroke_key="line_color", width_key="line_width", fill_key="inner_color") :
		# maybe this is too smart: fill=0 != fill=None
		if fill == None :
			attrs = '''
      <dia:attribute name="%s">
        <dia:color val="%s"/>
      </dia:attribute>
      <dia:attribute name="%s">
        <dia:real val="%.3f"/>
      </dia:attribute>
'''  % (stroke_key, self._rgb(color), width_key, self.line_width)
		elif fill :
			attrs = '''
      <dia:attribute name="%s">
        <dia:color val="%s"/>
      </dia:attribute>
      <dia:attribute name="show_background">
        <dia:boolean val="true"/>
      </dia:attribute>
      <!-- TODO: folding missing -->
      <dia:attribute name="%s">
        <dia:color val="%s"/>
      </dia:attribute>
      <dia:attribute name="%s">
        <dia:real val="0"/>
      </dia:attribute>
''' % (fill_key, self._rgb(color), stroke_key, self._rgb(color), width_key)
		else :
			attrs = '''
      <dia:attribute name="%s">
        <dia:color val="%s"/>
      </dia:attribute>
      <dia:attribute name="show_background">
        <dia:boolean val="false"/>
      </dia:attribute>
      <dia:attribute name="%s">
        <dia:real val="%.3f"/>
      </dia:attribute>
''' % (stroke_key, self._rgb(color), width_key, self.line_width)
		return attrs
	def _stroke_style (self) :
		return '''
      <dia:attribute name="line_style">
        <dia:enum val="%d"/>
      </dia:attribute>
      <dia:attribute name="dashlength">
        <dia:real val="%.3f"/>
      </dia:attribute>
''' % (self.line_style, self.dash_length)
	# just remember the given state to use it during object creation
	def set_linewidth (self, width) :
		self.line_width = width
	def set_linecaps (self, mode) :
		self.line_caps = mode
	def set_linejoin (self, mode) :
		self.line_join = mode
	def set_linestyle (self, style, length) :
		self.line_style = style
		self.dash_length = length
	def set_fillstyle (self, style) :
		# currently only 'solid' so not used anywhere else
		self.fill_style = style
	def set_font (self, font, size) :
		self.font = font
		self.font_size = size
	# now to the actual object creation - it should be 'delayed' to fold draw&fill again,
	# split in the rendering process but the Dia's 'Standard' objects can do both in one
	def _box (self, rect, color, fill) :
		self.oid = self.oid + 1
		self.f.write('''
    <dia:object type="Standard - Box" version="0" id="O%d">
      <dia:attribute name="elem_corner">
        <dia:point val="%.3f,%.3f"/>
      </dia:attribute>
      <dia:attribute name="elem_width">
        <dia:real val="%.3f"/>
      </dia:attribute>
      <dia:attribute name="elem_height">
        <dia:real val="%.3f"/>
      </dia:attribute>
%s%s
    </dia:object>''' % (self.oid, rect.left,rect.top, rect.right - rect.left, rect.bottom - rect.top,
			self._tinting(color, fill, 'border_color', 'border_width'), self._stroke_style()))
	def draw_rect (self, rect, fill, stroke) :
		if fill :
			self._box(rect, fill, 1)
		if stroke :
			self._box(rect, stroke, 0)

	def _ellipse (self, center, width, height, color, fill) :
		self.oid = self.oid + 1
		self.f.write('''
    <dia:object type="Standard - Ellipse" version="0" id="O%d">
      <dia:attribute name="elem_corner">
        <dia:point val="%.3f,%.3f"/>
      </dia:attribute>
      <dia:attribute name="elem_width">
        <dia:real val="%.3f"/>
      </dia:attribute>
      <dia:attribute name="elem_height">
        <dia:real val="%.3f"/>
      </dia:attribute>
%s%s
    </dia:object>''' % (self.oid, center.x-width/2, center.y-height/2, width, height,
			self._tinting(color, fill, 'border_color', 'border_width'), self._stroke_style()))
	def draw_ellipse (self, center, width, height, fill, stroke) :
		if fill :
			self._ellipse(center, width, height, fill, 1)
		if stroke :
			self._ellipse(center, width, height, stroke, 0)

	def _arc(self, center, width, height, angle1, angle2, color, fill) :
		self.oid = self.oid + 1
		mPi180 = math.pi / 180.0
		rx = width / 2.0
		ry = height / 2.0
		sx = center.x + rx * math.cos(mPi180 * angle1)
		sy = center.y - ry * math.sin(mPi180 * angle1)
		ex = center.x + rx * math.cos(mPi180 * angle2)
		ey = center.y - ry * math.sin(mPi180 * angle2)
		# the middle control point
		cx = center.x + rx * math.cos(mPi180 * (angle1 + angle2)/2.0)
		cy = center.y - ry * math.sin(mPi180 * (angle1 + angle2)/2.0)
		dx = (sx + ex) / 2.0 - cx
		dy = (sy + ey) / 2.0 - cy
		dist = math.sqrt(dx * dx + dy * dy)
		if angle1 > angle2 :
			dist = -dist
		self.f.write('''
    <dia:object type="Standard - Arc" version="0" id="O%d">
      <dia:attribute name="conn_endpoints">
        <dia:point val="%.3f,%.3f"/>
        <dia:point val="%.3f,%.3f"/>
      </dia:attribute>
      <dia:attribute name="curve_distance">
        <dia:real val="%.3f"/>
      </dia:attribute>
%s%s
    </dia:object>''' % (self.oid, sx, sy, ex, ey, dist,
			self._tinting(color, fill, 'arc_color'), self._stroke_style()))
	def draw_arc (self, center, width, height, angle1, angle2, color) :
		self._arc(center, width, height, angle1, angle2, color, None)
	def fill_arc (self, center, width, height, angle1, angle2, color) :
		self._arc(center, width, height, angle1, angle2, color, 1)

	def _poly (self, type_name, points, color, fill) :
		self.oid = self.oid + 1
		self.f.write('''
    <dia:object type="Standard - %s" version="0" id="O%d">
%s%s
      <dia:attribute name="poly_points">''' % (type_name, self.oid,
						self._tinting(color, fill), self._stroke_style()))
		for p in points :
			self.f.write('''
        <dia:point val="%.3f,%.3f"/>''' % (p.x, p.y))
		self.f.write('''
      </dia:attribute>
    </dia:object>''')
	def draw_polyline (self, points, color) :
		self._poly("PolyLine", points, color, None)
	def stroke_polygon (self, points, color) :
		self._poly("Polygon", points, color, 0)
	def fill_polygon (self, points, color) :
		self._poly("Polygon", points, color, 1)
	def draw_polygon (self, points, fill, stroke) :
		if fill :
			self.fill_polygon(points, fill)
		if stroke :
			self.stroke_polygon(points, fill)

	def _bezier (self, type_name, points, color, fill) :
		self.oid = self.oid + 1
		self.f.write('''
    <dia:object type="Standard - %s" version="0" id="O%d">
%s%s
      <dia:attribute name="bez_points">''' % (type_name, self.oid, self._tinting(color, fill), self._stroke_style()))
		for bp in points :
			if bp.type == 0 : # BEZ_MOVE_TO
				# moveto must be first (and only first) otherwise the file format is broken
				self.f.write('''
        <dia:point val="%.3f,%.3f"/>''' % (bp.p1.x, bp.p1.y))
			elif bp.type == 1 : # BEZ_LINE_TO
				# simulated by curveto (there is no lineto on the file format level)
				self.f.write('''
        <dia:point val="%.3f,%.3f"/>
        <dia:point val="%.3f,%.3f"/>
        <dia:point val="%.3f,%.3f"/>''' % (last.x, last.y, bp.p1.x, bp.p1.y, bp.p1.x, bp.p1.y))
			elif bp.type == 2 : # BEZ_CURVE_TO
				self.f.write('''
        <dia:point val="%.3f,%.3f"/>
        <dia:point val="%.3f,%.3f"/>
        <dia:point val="%.3f,%.3f"/>''' % (bp.p1.x, bp.p1.y, bp.p2.x, bp.p2.y, bp.p3.x, bp.p3.y))
			last = bp.p1
		self.f.write('''
      </dia:attribute>
    </dia:object>''')
	def draw_bezier (self, points, color) :
		self._bezier("BezierLine", points, color, None)
	def fill_bezier (self, points, color) :
		self._bezier("Beziergon", points, color, 1)

	def draw_string (self, text, pos, alignment, color) :
		self.oid = self.oid + 1
		self.f.write('''
    <dia:object type="Standard - Text" version="1" id="O%d">
      <dia:attribute name="obj_pos">
        <dia:point val="%.3f,%.3f"/>
      </dia:attribute>
      <dia:attribute name="text">
        <dia:composite type="text">
          <dia:attribute name="string">
            <dia:string>#%s#</dia:string>
          </dia:attribute>
          <dia:attribute name="font">
            <dia:font family="%s" style="%d"/>
          </dia:attribute>
          <dia:attribute name="height">
            <dia:real val="%.3f"/>
          </dia:attribute>
          <dia:attribute name="color">
            <dia:color val="%s"/>
          </dia:attribute>
          <dia:attribute name="alignment">
            <dia:enum val="%d"/>
          </dia:attribute>
        </dia:composite>
      </dia:attribute>
      <dia:attribute name="valign">
        <dia:enum val="3"/>
      </dia:attribute>
      </dia:object>''' % (self.oid, pos.x, pos.y, text.replace("&", "&amp;").replace(">", "&gt;").replace("<", "&lt;"),
			self.font.family, self.font.style, self.font_size,
			self._rgb(color), alignment))
	def draw_image (self, point, width, height, image) :
		fname = image.filename
		# do something better than absolute pathes
		common = os.path.commonprefix ([fname, self.filename])
		if len(common) > 0 and self.filename[len(common):].find(os.path.pathsep) == -1 :
			fname = fname[len(common):]
		# ensure <broken> does not get through
		fname = fname.replace("&", "&amp;").replace(">", "&gt;").replace("<", "&lt;")
		self.oid = self.oid + 1
		if abs(width/height - image.width/image.height) < 0.001 :
			keep_aspect = "true"
		else :
			keep_aspect = "false"
		self.f.write('''
    <dia:object type="Standard - Image" version="0" id="O%d">
      <dia:attribute name="elem_corner">
        <dia:point val="%.3f,%.3f"/>
      </dia:attribute>
      <dia:attribute name="elem_width">
        <dia:real val="%.3f"/>
      </dia:attribute>
      <dia:attribute name="elem_height">
        <dia:real val="%.3f"/>
      </dia:attribute>
      <dia:attribute name="file">
        <dia:string>#%s#</dia:string>
      </dia:attribute>
      <dia:attribute name="draw_border">
        <dia:boolean val="false"/>
      </dia:attribute>
      <dia:attribute name="keep_aspect">
        <dia:boolean val="%s"/>
      </dia:attribute>
    </dia:object>''' % (self.oid, point.x,point.y, width, height, fname, keep_aspect))

	# defining only this line drawer makes almost everything interpolated
	def draw_line (self, start, end, color) :
		self.oid = self.oid + 1
		self.f.write('''
    <dia:object type="Standard - Line" version="0" id="O%d">
      <dia:attribute name="conn_endpoints">
        <dia:point val="%.3f,%.3f"/>
        <dia:point val="%.3f,%.3f"/>
      </dia:attribute>
%s%s
    </dia:object>''' % (self.oid, start.x,start.y, end.x, end.y,
			self._tinting(color), self._stroke_style()))

# register the renderer
dia.register_export (_("Dia plain"), "dia", StandardDiaRenderer())




import sys, dia

import gettext
_ = gettext.gettext

# sys.path.insert(0, 'd:/graph/dia/dia')

def dia_debug_cb (data, flags) :
	"gets passed in the active diagram, flags are unused at the moment"
	for layer in data.layers :
		print("Layer :", layer.name)
		for o in layer.objects :
			print(str(o), str(o.bounding_box))

def dia_debug_props_cb (data, flags) :
	for layer in data.layers :
		print("Layer :", layer.name)
		for o in layer.objects :
			print(str(o))
			props = o.properties
			for s in list(props.keys()) :
				print(props[s].type + " " + s + " (visible=%d)" % props[s].visible)
				try :
					p = props[s].value
				except :
					p = None
				print("\t" + str(p))

# dia-python keeps a reference to the renderer class and uses it on demand
dia.register_action ("DebugBoundingbox", _("Dia _BoundingBox Debugger"),
                     "/DisplayMenu/Debug/DebugExtensionStart",
                     dia_debug_cb)

dia.register_action ("DebugProperty", _("Dia _Property API Debugger"),
                     "/DisplayMenu/Debug/DebugExtensionStart",
                     dia_debug_props_cb)



# PyDia DOT Import
# Copyright (c) 2009 Hans Breuer <hans@breuer.org>

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

##
# \file dot2dia.py \brief translate dot ( http://www.graphviz.org/ ) to Dia format
# \ingroup ImportFilters

import re, sys

import gettext
_ = gettext.gettext

# FIXME: keywords are case indepentend
keywords = ['node', 'edge', 'graph', 'digraph', 'strict']
# starts with either a keyword or a node name in quotes.
# BEWARE: (?<-> ) - negative lookbehind to not find nodes a second time in connection definition (and misinterpret the params)
rDecl = re.compile(r'\s*(?<!-> )(?P<cmd>(?:' + ')|(?:'.join(keywords) + ')|(?:\w+' + ')|(?:"[^"]+"))\s+\[(?P<dict>[^\]]+)\];', re.DOTALL | re.MULTILINE)
# dont assume that all node names are in quotes
rEdge = re.compile(r'\s*(?P<n1>("[^"]+")|(\w+))\s*->\s*(?P<n2>("[^"]+")|(\w+))\s+\[(?P<dict>[^\]]+)\];*', re.DOTALL | re.MULTILINE)
# a list of key=value
rParam = re.compile(r'(?P<key>\w+)\s*=(?P<val>(\w+)|("[^"]+")),?\s*', re.DOTALL | re.MULTILINE)

# units in dot are either points or inch
cmInch = 2.54
cmPoints = cmInch/72.0
# dot y up, dia y down

def StripQuotes(s) :
	"strip quotes if any"
	if s[0] == '"' :
		s = s[1:-1]
	return s

def DictFromString (s) :
	#print "KVs", s
	d = {}
	for m in rParam.finditer (s) :
		if m :
			d[m.group ("key")] = StripQuotes(m.group ("val"))
	return d

##
# \brief Accumulating information with _DiaObject
class Object :
	""" will end as a Dia Object """
	def __init__ (self, typename, parms) :
		self.typename = typename
		self.parms = parms
	def FontSize (self) :
		try :
			return float(self.parms['fontsize']) * cmPoints
		except :
			return 0.6
##
# \brief The nodes of the graph - finally represented as _Ellipse
class Node(Object) :
	def __init__ (self, name, parms) :
		Object.__init__(self, "Standard - Ellipse", parms)
		self.name = name
	def Pos(self) :
		'deliver scaled X,Y coordinate'
		x, y = 0.0, 0.0
		try :
			xy = self.parms['pos'].split(',')
			x = float(xy[0]) * cmPoints
			y = float(xy[1]) * cmPoints
		except :
			print("No position on '%s'" % (self.name,))
		return x,-y
	def Size(self) :
		'deliver scaled W,H coordinate'
		w, h = 0.5, 0.5
		try :
			w = float(self.parms['width']) * cmInch #? maybe this is relative to the font size?
			h = float(self.parms['height']) * cmInch
		except :
			print("No size on '%s'" % (self.name,))
		return w,h

##
# \brief The edges of the graph - finally represented as _Bezierline
class Edge(Object) :
	def __init__ (self, src, dest, parms) :
		Object.__init__(self, "Standard - BezierLine", parms)
		self.src = src
		self.dest = dest
	def LabelPos (self) :
		x, y = 0.0, 0.0
		try :
			xy = self.parms['lp'].split(',')
			x = float(xy[0]) * cmPoints
			y = float(xy[1]) * cmPoints
		except :
			if 'label' in self.parms :
				# should be optional otherwise
				print("No label pos on %s" % (self.src + '->' + self.dest,))
		return x, -y
	def Pos (self) :
		# no need to do something smart, it get adjusted anyway
		return 0.0, 0.0
	def SetPoints (self, diaobj) :
		'the property to set must match the type'
		pts = []
		if 'pos' in self.parms :
			s = self.parms['pos']
			if s[:2] == 'e,' :
				sp = s[2:].split(" ")
				# apparently the first point is the end? just put it there!
				sp.append(sp[-1])
				del sp[0]
				bp = []
				for i in range(0, len(sp)) :
					xy = sp[i].replace("\n", "").replace("\\", "").split(",")
					try :
						x = float(xy[0]) * cmPoints
						y = float(xy[1]) * (-cmPoints)
					except ValueError :
						print(xy)
						continue
					bp.append((x,y))
					# must convert to _one_ tuple
					if i == 0 : # first must be move to
						pts.append ((0, bp[0][0], bp[0][1]))
						bp = [] # reset to new point
					elif len(bp) == 3 : # rest is curveto ==2
						pts.append ((2, bp[0][0], bp[0][1], bp[1][0], bp[1][1], bp[2][0], bp[2][1]))
						bp = [] # reset
			if len(bp) > 0 :
				print(len(bp), "bezier points left!")
		if len(pts) > 1 :
			diaobj.properties['bez_points'] = pts
		else :
			print("BezPoints", pts)

def MergeParms (d, extra) :
	for k in list(extra.keys()) :
		if k not in d :
			d[k] = extra[k]

##
# \brief Parsing the given dot file
def Parse(sFile) :
	f = open(sFile, 'r')
	s = f.read()
	extra = {}
	nodes = {}
	edges = []
	if 0 : # debug regex
		dbg = rDecl.findall(s)
		for db in dbg :
			print(db)
	for m in rDecl.finditer(s) :
		if m :
			name = StripQuotes(m.group("cmd"))
			if name in keywords :
				if name in extra :
					MergeParms(extra[name], DictFromString(m.group("dict")))
				else :
					extra[name] = DictFromString(m.group("dict"))
			else : # must be a node
				n = Node(name, DictFromString(m.group("dict")))
				if 'node' in extra :
					MergeParms(n.parms, extra['node'])
				nodes[name] = n
	for m in rEdge.finditer(s) :
		if m :
			# the names given are not necessarily registered as nodes already
			defparams = {}
			if 'node' in extra :
				defparams = extra['node']
			for k in ["n1", "n2"] :
				name = StripQuotes(m.group(k))
				if name in nodes :
					pass # defparms should be set above
				else :
					nodes[name] = Node(name, defparams)
			# remember connection
			edges.append(Edge(StripQuotes(m.group("n1")), StripQuotes(m.group("n2")), DictFromString(m.group("dict"))))
	return [nodes, edges]

##
# \brief Adding a label for the edges
#
# This function could be improved if Dia would allow to
# attach labels to arbitrary objects. For the time being
# only the initial postion does match, but relayouting the
# graph in Dia will loose the position
def AddLabel (layer, pos, label, fontsize, center=0) :
	""" create a Text object an put it into the layer """
	textType = dia.get_object_type("Standard - Text")
	obj, h1, h2 = textType.create(pos[0], pos[1])
	#TODO: transfer font-size
	obj.properties["text"] = label
	obj.properties["text_height"] = fontsize
	if center :
		obj.properties["text_alignment"] = 1
		obj.properties["text_vert_alignment"] = 2
	layer.add_object(obj)

##
# \brief Callback registered for the ImportFilter
def ImportFile (sFile, diagramData) :
	""" read the dot file and create diagram objects """
	nodes, edges = Parse(sFile)
	layer = diagramData.active_layer # can do better, e.g. layer per graph
	for key in list(nodes.keys()) :
		n = nodes[key]
		nodeType = dia.get_object_type(n.typename) # could be optimized out of loop
		x, y = n.Pos()
		w, h = n.Size()
		obj, h1, h2 = nodeType.create(x-w/2, y-h/2) # Dot pos is center, Dia (usually) uses top/left
		# resizing the Ellipse by handle is screwed
		# obj.move_handle(h2, (x+w/2, y+h/2), 0, 0) # resize the object
		obj.properties["elem_width"] = w
		obj.properties["elem_height"] = h
		if 'fillcolor' in n.parms :
			try :
				obj.properties['fill_colour'] = n.parms['fillcolor'] # same color syntax?
			except :
				print("Failed to apply:", n.parms['fillcolor'])
		layer.add_object(obj)
		AddLabel (layer, (x,y), n.name, n.FontSize(), 1)
		obj.properties['meta'] = n.parms # copy all (remaining) parameters
		# after creation replace the node with the object (needed to connect them)
		nodes[key] = obj
	for e in edges :
		edgeType = dia.get_object_type(e.typename) # could be optimized out of loop
		x, y = e.Pos() # just to have a start
		con, h1, h2 = edgeType.create(x,y)
		e.SetPoints(con)
		if 'style' in e.parms : # set line style
			con.properties['line_style'] = (4, 0.5) #FIXME: hard-coded dotted
		if 'weight' in e.parms :
			con.properties['line_width'] = float(e.parms['weight']) / 10.0 # arbitray anyway
		layer.add_object(con)
		if e.src in nodes :
			h = con.handles[0]
			obj = nodes[e.src]
			# by moving to the cp position first, the connection's points get recalculated
			pos = obj.connections[8].pos
			con.move_handle(h, pos, 0, 0)
			h.connect(obj.connections[8]) # connect to mid-point
		if e.dest in nodes :
			h = con.handles[-1]
			obj = nodes[e.dest]
			pos = obj.connections[8].pos
			con.move_handle(h, pos, 0, 0)
			h.connect (obj.connections[8]) # connect to mid-point
		if 'label' in e.parms :
			AddLabel (layer, e.LabelPos(), e.parms['label'], e.FontSize())
	diagram = None # FIXME: get it
	if diagram :
		for n, o in nodes.items() :
			diagram.update_connections(o)
		diagram.update_extents()
	return diagramData

if __name__ == '__main__':
	# just testing at the moment
	nodes, edges = Parse(sys.argv[1])
	for k, n in nodes.items() :
		print("Name:", n.name, "Pos:", n.Pos(), "WxH:", n.Size())
	for e in edges :
		print(e.src, "->", e.dest, e.LabelPos(), e.parms)
else :
	# run as a Dia plug-in
	import dia
	dia.register_import (_("Graphviz Dot"), "dot", ImportFile)




def read_cgm(file):
	if type(file) == type(''):
		file = open(file, 'rb');
	list = []
	head = file.read(2);
	while head != '':
		if len(head) != 2: break
		el_class = ord(head[0]) >> 4
		el_id = ((ord(head[0]) & 0x0f) << 3) | (ord(head[1]) >> 5)
		el_len = ord(head[1]) & 0x1f
		if el_len == 31:  # long form element
			head = file.read(2)
			if len(head) != 2: break
			el_len = (ord(head[0]) << 8) | ord(head[1])
		print((el_class, el_id, el_len))

		el_body = file.read(el_len)
		if el_len & 1 == 1:  # odd element length
			file.read(1) # discard NULL byte

		list.append((el_class, el_id, el_body))

		if (el_class, el_id) == (0, 2):
			break
		
		head = file.read(2)
	return list



#    PyDia imgmap.py : produce an html image map
#    Copyright (c) 2007  Hans Breuer  <hans@breuer.org>

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import dia, sys, os.path

import gettext
_ = gettext.gettext

class ObjRenderer :
	def __init__ (self) :
		self.f = None
		# the bintmap exporters calculate their ow margins from the bunding box
		self.xofs = 0
		self.yofs = 0
	def begin_render (self, data, filename) :
		self.f = open(filename, "w")
		bMapLayer = 0 # if there is a special layer containing the links
		for layer in data.layers :
			if layer.name == "map" :
				bMapLayer = 1
				break
		scale = 20.0 * data.paper.scaling
		r = data.extents
		width = int((r.right - r.left) * scale)
		height = int((r.bottom - r.top) * scale)
		name = os.path.split (filename)[1]
		fname = name[:name.find(".")] + ".png" # guessing
		self.f.write ('<image src="%s" width="%d", height="%d" usemap="#%s">\n' % (fname, width, height, name))
		self.f.write ('<map name="%s">\n' % (name,))

		self.xofs = - (r.left * scale)
		self.yofs = - (r.top * scale)
		if bMapLayer :
			self.WriteAreas (layer, scale)
		else :
			for layer in data.layers :
				self.WriteAreas (layer, scale)
		self.f.write ('</map>\n')

	def WriteAreas (self, layer, scale) :
		for o in layer.objects :
			r = o.bounding_box
			if "name" in o.properties :
				url = o.properties["name"].value
			elif "text" in o.properties :
				url = o.properties["text"].value.text
			else :
				continue
			if len(url) == 0 or url.find (" ") >= 0 :
				continue

			alt = url
			if "comment" in o.properties :
				alt = o.properties["comment"].value
			# need to sale the original coords to the bitmap size
			x1 = int(r.left * scale) + self.xofs
			y1 = int(r.top * scale) + self.yofs
			x2 = int(r.right * scale) + self.xofs
			y2 = int(r.bottom * scale) + self.yofs
			self.f.write ('  <area shape="rect" href="%s" title="%s" alt="%s" coords="%d,%d,%d,%d">\n' % \
				 ("#" + url, url, alt, x1, y1, x2, y2))
	def end_render (self) :
		self.f.close()

# dia-python keeps a reference to the renderer class and uses it on demand
dia.register_export (_("Imagemap"), "cmap", ObjRenderer())



#    Select empyt objects
#    Copyright (c) 2008, Hans Breuer <hans@breuer.org>

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import dia

import gettext
_ = gettext.gettext

def select_empty_cb (data, flags) :
	objs = data.active_layer.objects
	for o in objs :
		if o.bounding_box.right == o.bounding_box.left \
			or o.bounding_box.top == o.bounding_box.bottom :
			data.select (o)

dia.register_action ("SelectEmpty", _("_Empty"),
                       "/DisplayMenu/Select/By/SelectByExtensionStart",
                       select_empty_cb)



#  autolayoutforce.py - graph layout plug-in for Dia
#
#  Copyright (C) 2008 Frederic-Gerald Morcos <fred.mrocos@gmail.com>
#  Copyright (c) 2008 Hans Breuer <hans@breuer.org>
#
#  Playground for the "force based autolayout" algorithm initially implemented
# for Dia in C by Fred Morcos
# See also: http://en.wikipedia.org/wiki/Force-based_algorithms

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
import math, dia

import gettext
_ = gettext.gettext

def bbox_area (o) :
	r = o.bounding_box
	return (r.bottom - r.top) * (r.right - r.left)

def attraction (aconst, node, other) :
	"Calculates the attraction between two connected elements"
	p1 = node.properties["obj_pos"].value
	p2 = other.properties["obj_pos"].value
	x = (p2.x - p1.x) * aconst
	y = (p2.y - p1.y) * aconst
	return (x,y)

def repulsion (rconst, node, other) :
	"Calculates the repulsion between any two elements"
	p1 = node.properties["obj_pos"].value
	p2 = other.properties["obj_pos"].value

	m2 = bbox_area (other)
	m1 = bbox_area (node)

	dx = p1.x - p2.x
	dy = p1.y - p2.y

	denom = math.pow (dx * dx + dy * dy, 1.5) # magic number?
	numer = m1 * m2 * rconst

	try :
		fx = (numer * dx) / denom
		fy = (numer * dy) / denom
	except ZeroDivisionError :
		print("ZeroDivisionError")
		return (0,0)

	return (fx,fy)

def layout_force (nodes, rconst, aconst, timestep, damping) :
	energy = [0.0, 0.0]
	for o in nodes :
		netforce = [0.0, 0.0]
		velocity = [0.0, 0.0]
		for oo in nodes :
			if oo != o :
				r = repulsion (rconst, o, oo)
				netforce[0] += r[0]
				netforce[1] += r[1]
		for cpt in o.connections : # connection points
			for co in cpt.connected : # connected objects in this point
				edge = co
				oo = None
				for h in edge.handles : # the edge handles are connected to ...
					cto = h.connected_to # ... the other objects connection point
					if  cto and cto.object != o : # ... one of it being the current
						oo = cto.object
						# we usually only find _one_ other handle but there are exception
						# e.g. "Network - Bus"
						a = attraction (aconst, o, oo)
						netforce[0] += a[0]
						netforce[1] += a[1]
		#print "Netforce", netforce, timestep, damping
		velocity[0] = timestep * netforce[0] * damping
		velocity[1] = timestep * netforce[1] * damping

		# new position
		p = o.properties["obj_pos"].value
		px = p.x + timestep * velocity[0]
		py = p.y + timestep * velocity[1]
		#print "move", timestep * velocity[0], timestep * velocity[1]
		o.move (px, py)

		mass = bbox_area (o)
		energy[0] += mass * math.pow (velocity[0], 2) / 2
		energy[1] += mass * math.pow (velocity[1], 2) / 2
	return energy

def layout_force_cb(data, flags):
	# the things (nodes) we are moving around are all connected 'elements',
	# connection objects are only moving as a side effect
	nodes = []
	for o in data.selected :
		for cpt in o.connections : # ConnectionPoint
			if len (cpt.connected) > 0 :
				nodes.append (o)
				break
	# this ususally is an iterative process, finished if no energy is left
	#FIXME: layout_force (nodes, 2.0, 2.0, 0.2, 0.5) PASSES 0.0, 0.0
	e = layout_force (nodes, 2.0, 3.0, 2e-1, 5e-1)
	n = 0 # arbitrary limit to avoid endless loop
	while (e[0] > 1 or e[1] > 1) and n < 100 :
		e = layout_force (nodes, 2.0, 3.0, 2e-1, 5e-1)
		n += 1
	for o in nodes :
		data.update_connections (o)
	data.active_layer.update_extents() # data/diagram _update_extents don't recalculate?
	data.update_extents ()
	data.flush()
	print(n, "iterations")

dia.register_action ("LayoutForcePy", _("_Layout (force)"),
                     "/DisplayMenu/Test/TestExtensionStart",
                     layout_force_cb)



# PyDia Code Generation from UML Diagram
# Copyright (c) 2005  Hans Breuer <hans@breuer.org>

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, dia

import gettext
_ = gettext.gettext

class Klass :
	def __init__ (self, name) :
		self.name = name
		# use a list to preserve the order
		self.attributes = []
		# a list, as java/c++ support multiple methods with the same name
		self.operations = []
		self.comment = ""
		self.parents = []
		self.templates = []
		self.inheritance_type = ""
	def AddAttribute(self, name, type, visibility, value, comment, class_scope) :
		self.attributes.append ((name, (type, visibility, value, comment, class_scope)))
	def AddOperation(self, name, type, visibility, params, inheritance_type, comment, class_scope) :
		self.operations.append((name,(type, visibility, params, inheritance_type, comment, class_scope)))
	def SetComment(self, s) :
		self.comment = s
	def AddParrent(self, parrent):
		self.parents.append(parrent)
	def AddTemplate(self, template):
		self.templates.append(template)
	def SetInheritance_type(self, inheritance_type):
		self.inheritance_type = inheritance_type

##
# \brief Base class of all the code generators
# \extends _DiaPyRenderer
# \ingroup PyDia
class ObjRenderer :
	"Implements the Object Renderer Interface and transforms diagram into its internal representation"
	def __init__ (self) :
		# an empty dictionary of classes
		self.klasses = {}
		self.arrows = []
		self.filename = ""

	def begin_render (self, data, filename) :
		self.filename = filename
		# not only reset the filename but also the other state, otherwise we would accumulate information through every export
		self.klasses = {}
		self.arrows = []
		for layer in data.layers :
			# for the moment ignore layer info. But we could use this to spread accross different files
			for o in layer.objects :
				if o.type.name == "UML - Class" :
					#print o.properties["name"].value
					k = Klass (o.properties["name"].value)
					k.SetComment(o.properties["comment"].value)
					if o.properties["abstract"].value:
						k.SetInheritance_type("abstract")
					if o.properties["template"].value:
						k.SetInheritance_type("template")
					for op in o.properties["operations"].value :
						# op : a tuple with fixed placing, see: objects/UML/umloperations.c:umloperation_props
						# (name, type, comment, stereotype, visibility, inheritance_type, class_scope, params)
						params = []
						for par in op[8] :
							# par : again fixed placement, see objects/UML/umlparameter.c:umlparameter_props
							# (name, type, value, comment, kind)
							params.append((par[0], par[1], par[2], par[3], par[4]))
						k.AddOperation (op[0], op[1], op[4], params, op[5], op[2], op[7])
					#print o.properties["attributes"].value
					for attr in o.properties["attributes"].value :
						# see objects/UML/umlattributes.c:umlattribute_props
						#print "\t", attr[0], attr[1], attr[4]
						# name, type, value, comment, visibility, abstract, class_scope
						k.AddAttribute(attr[0], attr[1], attr[4], attr[2], attr[3], attr[6])
					self.klasses[o.properties["name"].value] = k
					#Connections
				elif o.type.name == "UML - Association" :
					# should already have got attributes relation by names
					pass
				# other UML objects which may be interesting
				# UML - Note, UML - LargePackage, UML - SmallPackage, UML - Dependency, ...

		edges = {}
		for layer in data.layers :
			for o in layer.objects :
				for c in o.connections:
					for n in c.connected:
						if not n.type.name in ("UML - Generalization", "UML - Realizes"):
							continue
						if str(n) in edges:
							continue
						edges[str(n)] = None
						if not (n.handles[0].connected_to and n.handles[1].connected_to):
							continue
						par = n.handles[0].connected_to.object
						chi = n.handles[1].connected_to.object
						if not par.type.name == "UML - Class" and chi.type.name == "UML - Class":
							continue
						par_name = par.properties["name"].value
						chi_name = chi.properties["name"].value
						if n.type.name == "UML - Generalization":
							self.klasses[chi_name].AddParrent(par_name)
						else: self.klasses[chi_name].AddTemplate(par_name)

	def end_render(self) :
		# without this we would accumulate info from every pass
		self.attributes = []
		self.operations = []

##
# \brief Generate a Python source file from an UML class diagram
class PyRenderer(ObjRenderer) :
	def __init__(self) :
		ObjRenderer.__init__(self)
	def end_render(self) :
		f = open(self.filename, "w")
		for sk in list(self.klasses.keys()) :
			parents = self.klasses[sk].parents + self.klasses[sk].templates
			if not parents:
				f.write ("class %s :\n" % (sk,))
			else:
				f.write ("class %s (%s) :\n" % (sk,", ".join(parents)))
			k = self.klasses[sk]
			if len(k.comment) > 0 :
				f.write ("\t'''" + k.comment + "'''\n")
			f.write ("\tdef __init__(self) :\n")
			for sa, attr in k.attributes :
				value = attr[2] == "" and "None" or attr[2]
				f.write("\t\tself.%s = %s # %s\n" % (sa, value, attr[0]))
			else :
				f.write("\t\tpass\n")
			for so, op in k.operations :
				# we only need the parameter names
				pars = "self"
				for p in op[2] :
					pars = pars + ", " + p[0]
				f.write("\tdef %s (%s) :\n" % (so, pars))
				if op[4]: f.write("\t\t\"\"\" %s \"\"\"\n" % op[4])
				f.write("\t\t# returns %s\n" % (op[0], ))
				f.write("\t\tpass\n")
		f.close()
		ObjRenderer.end_render(self)

##
# \brief Generate C++ source files from an UML class diagram
class CxxRenderer(ObjRenderer) :
	def __init__(self) :
		ObjRenderer.__init__(self)
	def end_render(self) :
		f = open(self.filename, "w")
		f.write("/* generated by dia/codegen.py */\n")
		# declaration
		for sk in list(self.klasses.keys()) :
			k = self.klasses[sk]
			if len(k.comment) > 0 :
				f.write ("/*" + k.comment + "*/\n")
			if len(k.parents) > 0 :
				f.write ("class %s : %s \n{\n" % (sk, ", ".join(k.parents)))
			else :
				f.write ("class %s \n{\n" % (sk,))
			# first sort by visibility
			ops = [[], [], [], []]
			for so, (t, v, p, i, c, s) in k.operations :
				ops[v].append((t,so,p,i,c))
			vars = [[], [], [], []]
			for sa, (t, vi, va, vc, class_scope) in k.attributes :
				#TODO: use 'va'=value 'vc'=comment
				vars[vi].append((t, sa))
			visibilities = ("public:", "private:", "protected:", "/* implementation: */")
			for v in [0,2,1,3] :
				if len(ops[v]) == 0 and len(vars[v]) == 0 :
					continue
				f.write ("%s\n" % visibilities[v])
				for op in ops[v] :
					if op[4] != "" :
						f.write ('\t/* ' + op[4] + ' */\n')
					inh = ""
					if op[3] < 2 :
						inh = "virtual "
					# detect ctor/dtor
					so = ""
					if sk == op[1] or ("~" + sk) == op[1] :
						so = "\t%s%s (" % (inh, op[1])
					else :
						so = "\t%s%s %s (" % (inh, op[0], op[1])
					f.write (so)
					# align parameters with the opening brace
					n = len(so)
					i = 0
					m = len(op[2]) - 1
					for p in op[2] :
						linefeed = ",\n\t" + " " * (n - 1)
						if i == m :
							linefeed = ""
						f.write ("%s %s%s" % (p[1], p[0], linefeed))
						i = i + 1
					if op[3] == 0 : # abstract ?= pure virtual?
						f.write(") = 0;\n")
					else :
						f.write(");\n")
				for var in vars[v] :
					f.write("\t%s %s;\n" % (var[0], var[1]))
			f.write ("};\n\n")
		# implementation
		# ...
		f.close()
		ObjRenderer.end_render(self)

##
# JsRenderer: export Dia UML diagram to Object style ECMA5 javascript code.
#
# Features:
#  - All methods and attributes are public.
#  - Inheriting from one class.
#  - Comments for classes, attributes and operations are supported.
#  - Method parameter default values are supported.
#  - Order of attributes and operations is preserved.
#  - Auto generate type check.
#
class JsRenderer(ObjRenderer) :
	def __init__(self) :
		ObjRenderer.__init__(self)
	def end_render(self) :
		f = open(self.filename, "w")
		f.write("/* generated by dia/codegen.py */\n\n")
		for sk in list(self.klasses.keys()) :
			f.write("// class definition : %s\n" % (sk, ))
			parents = self.klasses[sk].parents
			#add inheritance
			if parents:
				f.write("// Inherits from %s.\n" % (parents[0], ))
				f.write("%s.prototype = new %s();\n" % (sk, parents[0]))
				f.write("%s.prototype.constructor = %s;\n\n" % (sk, sk))
			f.write ("function %s () {\n" % (sk,))
			k = self.klasses[sk]
			if len(k.comment) > 0 :
				f.write ("\n\t/*" + k.comment + "*/\n")
			#use strict
			f.write("\t\"use strict\";\n")
			if k.attributes:
				f.write("\n")
			#all attributes are public
			for sa, attr in k.attributes :
				value = attr[2] == "" and "null" or attr[2]
				f.write("\tthis.%s = %s;\n" % (sa, value))
				#expected types checks (constructor)
				if attr[0]:
					f.write("\tif (typeof(this.%s) !== '%s') {\n" % (sa, attr[0]));
					f.write("\t\tthrow \"Unexpected type '\"+typeof(this.%s)+\"' for attribute this.%s.\";\n" % (sa, sa))
					f.write("\t}\n")
			f.write("}\n\n")
			#operations are defined out of class definition
			#return the first item of given p (tuple) ie : parameter name
			def p_name(p):
				return p[0]
			for so, op in k.operations :
				# only parameter names needed
				pars = ", ".join(map(p_name,op[2]))
				f.write("%s.prototype.%s = function(%s) {\n" % (sk, so, pars))
				if op[4]: f.write("\t/* %s */\n" % op[4])
				f.write("\t\"use strict\";\n\n")
				for p in op[2]:
					#default values :
					if len(p[2]) > 0 :
						t = (p[0], p[0], p[2])
						f.write("\tif (%s === undefined) {\n\t%s = %s;\n\t}\n" % t)
					#expected types checks
					if len(p[3]) > 0 :
						f.write("\tif (typeof(%s) !== '%s') {\n" % (p[0], p[3]));
						f.write("\t\tthrow \"Unexpected type '\"+typeof(%s)+\"' for parameter %s.\";\n" % (p[0], p[0]))
						f.write("\t}\n");
				f.write("\t// returns %s\n" % (op[0], ))
				f.write("\treturn ;\n")
				f.write("};\n\n")
		f.close()
		ObjRenderer.end_render(self)

##
# PascalRenderer: export Dia UML diagram to Object Pascal (Free Pascal, Delphi)
#
# Please follow some "drawing guidelines" and "naming conventions" so that the
# exporter can do its job.
#  - Use "UML - Generalization" arrows for class inheritance.
#  - Use "UML - Realizes" arrows when implementing an interface.
#  - Set a class to be "abstract" to denote it is an interface definition.
#  - Set Inheritance Type to "abstract" for 'virtual; abstract;' methods, set
#    it to "virtual" for 'virtual;' methods.
#  - Array fields are automatically recognized. If the name ends with "[]"
#    an 'Array of' is used. If the name uses a number with "[1234]", an
#    'Array[0..1233] of' is used. If the name uses a constant with
#    "[MaxEntries]" an 'Array[0..MaxEntries-1] of' is written.
#  - To inherit from classes which are not drawn (e.g. LCL/VCL classes),
#    name then class with the parent class in paranthesis (e.g.
#    "TMainWin(TForm)"
#
# Features
#  - Inheriting from one class and implementing multiple interfaces is
#    supported.
#  - Comments for classes, attributes and operations are supported. They are
#    put in the line before the method declaration with a '///' style comment
#    (Doxygen-like).
#  - Method parameter directions are supported (-> 'Var', 'Out').
#  - Method parameter default values are supported.
#  - 'Function' and 'Procedure' are automatically recognized by whether a
#    return type exists or not.
#  - order of classes is alphabetically
#  - the order of attributes and operations is preserved
#  - Prints a list of forward declarations of all classes at the beginning
#    to avoid declaration order problems.
#
# TODO:
#  - Automatically use the keyword "override" instead of "virtual" in
#    descendant classes.
#  - Automatically define 'Properties'. Unfortunately the UML standard
#    doesn't support this and so the Dia dialog has no option to specify
#    this. So a "code" has to be used.
#  - Mark/recognize Constructors and Destructors
#  - Write comments for method parameters (e.g. by using a big doxygen
#    comment '(** ... *)' before the method)
#  - Use "Packages" to split the classes in separate 'Unit's.
#  - Beautify and comment the export code. Using arrays with "magic number
#    indexes" for certain fields is bad and tedious to work with.
#  - Support defining global constants.
#  - Support defining global types (especially for enums, arrays,
#    records, ...).
#  - Apply some sanity checks to the UML diagram:
#     - multiple inheritance is forbidded
#     - if implementing an interface, all required methods must be
#       implemented; alternative: just put all methods there, so the UML
#       drawer doesn't have to write them twice
#     - visibility for all methods of an interfaces must be "public"
#     - don't write the visibility specifier 'public' for interfaces
#     - no "Attributes" for interface definitions, but properties are
#       allowed
#     - default values for method parameters must be the last parameters
class PascalRenderer(ObjRenderer) :
	def __init__(self) :
		ObjRenderer.__init__(self)
	def end_render(self) :
		f = open(self.filename, "w")
		f.write("/* generated by dia/codegen.py */\n")
		f.write("Type\n")
		# classes
		class_names = list(self.klasses.keys())
		class_names.sort()
		# forward declarations of all classes
		for sk in class_names :
			k = self.klasses[sk]
			# class declaration
			if k.inheritance_type == "abstract" :
				f.write ("  %s = interface;\n" % (sk))
			else :
				f.write ("  %s = class;\n" % (sk))
		f.write("\n");
		# class declarations
		for sk in class_names :
			k = self.klasses[sk]
			# comment
			if len(k.comment) > 0 :
				f.write("  /// %s\n" % (k.comment))
			# class declaration
			if k.inheritance_type == "abstract" :
				f.write ("  %s = interface" % (sk))
			else :
				f.write ("  %s = class %s" % (sk, k.inheritance_type))
			# inherited classes / implemented interfaces
			p = []
			if k.parents :
				p.append(k.parents[0])
			if k.templates :
				p.append(",".join(k.templates))
			if len(p) > 0 :
				f.write("(%s)" % ",".join(p))
			f.write ("\n")
			# first sort by visibility
			ops = [[], [], [], [], [], []]
			for op_name, (op_type, op_visibility, op_params, op_inheritance, op_comment, op_class_scope) in k.operations :
				ops[op_visibility].append((op_type, op_name, op_params, op_comment, op_inheritance))
			vars = [[], [], [], []]
			for var_name, (var_type, var_visibility, var_value, var_comment, class_scope) in k.attributes :   # name, type, visibility, value, comment
				vars[var_visibility].append((var_type, var_name, var_value, var_comment))
			visibilities = ("public", "private", "protected", "/* implementation */")
			for v in [1,2,0,3] :
				if len(ops[v]) == 0 and len(vars[v]) == 0 :
					continue
				# visibility prefix
				f.write ("  %s\n" % visibilities[v])
				# variables
				for var in vars[v] :
					# comment
					if len(var[3]) > 0 :
						f.write ("    /// %s\n" % var[3])
					if var[1].endswith("]") :
						# array; check if this is dynamic or with defined size
						i = var[1].find("[")
						varname = var[1]
						arraysize = varname[i+1:-1]
						varname = varname[:i]
						if len(arraysize) > 0 :
							# array with defined size
							if arraysize.find("..") > 0 :
								f.write("    %s : Array[%s] of %s;\n" % (varname, arraysize, var[0]))
							elif arraysize.isdigit() :
								arraysize = int(arraysize)-1
								f.write("    %s : Array[0..%d] of %s;\n" % (varname, arraysize, var[0]))
							else :
								f.write("    %s : Array[0..%s-1] of %s;\n" % (varname, arraysize, var[0]))
						else :
							# dynamic size
							f.write("    %s : Array of %s;\n" % (varname, var[0]))
					else :
						# normal variable
						f.write("    %s : %s;\n" % (var[1], var[0]))
				# operations
				for op in ops[v] :
					if len(op[3]) > 0 :
						f.write ("    /// %s\n" % op[3])
					if len(op[0]) == 0 :
						f.write ("    Procedure %s" % op[1])
					else :
						f.write ("    Function %s" % op[1])
					if len(op[2]) > 0 :
						f.write ("(")
						i = 0
						m = len(op[2]) - 1
						for p in op[2] :
							if p[4] == 2 :
								f.write ("Out ")
							elif p[4] == 3 :
								f.write ("Var ")
							f.write ("%s:%s" % (p[0], p[1]))
							if len(p[2]) > 0 :
								f.write (":=%s" % p[2])
							if i != m :
								f.write(";")
							i = i + 1
						f.write (")")
					if len(op[0]) == 0 :
						f.write(";")
					else :
						f.write (" : %s;" % op[0])
					# inheritance type
					if op[4] == 0 :
						f.write (" virtual; abstract;");
					elif op[4] == 1 :
						f.write (" virtual;");
					f.write ("\n")
			f.write ("  End;\n\n")
		# implementation
		# ...
		f.close()
		ObjRenderer.end_render(self)

##
# JavaRenderer: export Dia UML diagram to Java
#
# improved by: Manuel Arguelles
#
# Features:
#	* Comments for classes, attributes, methods and methods parameters are
#	  supported. Uses a comment style javaDoc-like. The comments are
#	  divided into lines if they exceed 79 characters.
#	* Splits the classes in separate files. Creates a file that contains
#	  the dia/codegen.py firm :)
#
# Fixes:
#	* Visibilities "private" and "protected" were reversed.
#	* Comments for classes, attributes, methods and methods parameters.
#	* Comments are divided into lines if they exceed 79 characters.
#	* Splits the classes in separate files.
#	* Not write "NULL" in empty comments.
#	* Writes the default values of the arguments of the methods.
#	* Makes public all the classes and interfaces.
#	* Not write "static" in all the methods.
#
class JavaRenderer(ObjRenderer) :
	def __init__(self) :
		ObjRenderer.__init__(self)

	def end_render(self) :
		visibilities = {0:"public", 1:"private", 2:"protected"}

		mainfile = open(self.filename, "w")
		mainfile.write("/* Generated by dia/codegen.py\n *\n * Generated files:\n")
		for name, klass in self.klasses.items() :
			# splits the classes in separate files
			classfile = self.filename[:self.filename.rfind("/")+1] + name.capitalize() + ".java"
			f = open(classfile, "w")
			# class comment
			f.write("/**\n")
			if len(klass.comment) > 1:
				for l in range(len(klass.comment)/77+1):
					f.write(" * %s\n" % klass.comment[l*77:(l+1)*77])
			f.write(" * @author\n *\n */\n")

			if klass.inheritance_type == "template": classtype = "public interface"
			elif klass.inheritance_type == "abstract": classtype = "public abstract class"
			else: classtype = "public class"
			f.write ("%s %s" % (classtype, name))
			if klass.parents:
				f.write (" extends %s" % klass.parents[0])
			if klass.templates:
				f.write (" implements %s" % ", ".join(klass.templates))
			f.write(" {\n")

			# attributes
			for attrname, (type, visibility, value, comment, class_scope) in klass.attributes :
				if comment:
					f.write("\t/**\n")
					for l in range(len(comment)/73+1):
						f.write("\t * %s\n" % comment[l*73:(l+1)*73])
					f.write("\t */\n")
				if visibility in visibilities:
					vis = visibilities[visibility]+" "
				else: vis = ""
				f.write("\t%s%s %s" % (vis, type, attrname))
				if value: f.write(" = %s" % value)
				f.write(";\n")

			# dafault constructor
			if not klass.inheritance_type == "template":
				f.write ("\n\tpublic %s() {\n\t\t\n\t}\n\n" % name)

			# We should automatic implement abstract parrent and interface methods
			parmethods = []
			if klass.parents:
				parmethods = [(n,m[:3]+(1,)+m[4:]) for n,m in \
						self.klasses[klass.parents[0]].operations if m[3] == 0]
			for template in klass.templates:
				parmethods.extend(self.klasses[template].operations)

			for pName, pMethod in parmethods:
				pTypes = [p[1] for p in pMethod[2]]
				for name, pars in [(n,m[2]) for n,m in klass.operations]:
					types = [p[1] for p in pars]
					if pars == pMethod[2] and types == pTypes:
						break
				else: klass.operations.append((pName,pMethod))

			for methodname, method in klass.operations :
				# method comment
				f.write("\t/**\n")
				if method[4]:
					for l in range(len(method[4])/73+1):
						f.write("\t * %s\n" % method[4][l*73:(l+1)*73])
				if method[2]:
					for param in method[2]:
						f.write("\t * @param %s\n" % param[0]) #param[0]->name
				f.write("\t * @return %s\n\t */\n" % (method[0]=="" and "void" or method[0]))
				# if there are no parameter names, something else should appear
				pars = []
				v = ord("a")
				for name, type, value, comment, kind in method[2]:
					#TODO: also use: kind
					if not name:
						name = chr(v)
						v += 1
					if value:
						value = "=" + value
					pars.append((type, name, value))
				pars = ", ".join([type+" "+name+value for type, name, value in pars])

				vis = method[1] in visibilities and visibilities[method[1]] or ""
				return_type = method[0] == "" and "void" or method[0]
				inheritance_type = method[3] == 0 and "abstract " or ""
				# this is wrong!
				#static = method[4] and "static " or ""
				static = ""
				f.write("\t%s %s%s%s %s(%s)" % (vis, static, inheritance_type, \
												return_type, methodname, pars))
				if klass.inheritance_type == "template" or method[3] == 0:
					f.write(";\n\n")
				else:
					f.write(" {\n\t\t// TODO Auto-generated method stub\n\t}\n\n")
			f.write ("}\n\n")
			f.close()
			mainfile.write(" *\t%s\n" % classfile)
		mainfile.write(" */\n")
		mainfile.close()
		ObjRenderer.end_render(self)

##
# PhpRenderer: export Dia UML diagram to PHP
#
# Added by: Steven Garcia
#
# This is similar to the Java renderer except for PHP
# Added class scope (static) support
class PhpRenderer(ObjRenderer) :
	def __init__(self) :
		ObjRenderer.__init__(self)

	def end_render(self) :
		visibilities = {0:"public", 1:"private", 2:"protected"}

		mainfile = open(self.filename, "w")
		mainfile.write("<?php\n/* Generated by dia/codegen.py\n *\n * Generated files:\n")

		for name, klass in self.klasses.items() :
			# splits the classes in separate files
			classfile = self.filename[:self.filename.rfind("/")+1] + name + ".php"
			f = open(classfile, "w")
			# class comment
			f.write("<?php\n/**\n")
			if len(klass.comment) > 1:
				for l in range(len(klass.comment)/77+1):
					f.write(" * %s\n" % klass.comment[l*77:(l+1)*77])
			f.write(" * @author\n *\n */\n")

			if klass.inheritance_type == "template": classtype = "interface"
			elif klass.inheritance_type == "abstract": classtype = "abstract class"
			else: classtype = "class"
			f.write ("%s %s" % (classtype, name))
			if klass.parents:
				f.write (" extends %s" % klass.parents[0])
			if klass.templates:
				f.write (" implements %s" % ", ".join(klass.templates))
			f.write("\n{\n")

			# attributes
			for attrname, (type, visibility, value, comment, class_scope) in klass.attributes :
				if comment:
					f.write("\t/**\n")
					for l in range(len(comment)/73+1):
						f.write("\t * %s\n" % comment[l*73:(l+1)*73])
					f.write("\t */\n")
				if visibility in visibilities:
					vis = visibilities[visibility]
				else: vis = ""
				static = class_scope and "static " or ""
				f.write("\t%s%s $%s" % (static, vis, attrname))
				if value: f.write(" = %s" % value)
				f.write(";\n")

			# We should automatic implement abstract parent and interface methods
			parmethods = []
			if klass.parents:
				parmethods = [(n,m[:3]+(1,)+m[4:]) for n,m in \
						self.klasses[klass.parents[0]].operations if m[3] == 0]
			for template in klass.templates:
				parmethods.extend(self.klasses[template].operations)

			for pName, pMethod in parmethods:
				pTypes = [p[1] for p in pMethod[2]]
				for name, pars in [(n,m[2]) for n,m in klass.operations]:
					types = [p[1] for p in pars]
					if pars == pMethod[2] and types == pTypes:
						break
				else: klass.operations.append((pName,pMethod))

			for methodname, method in klass.operations :
				# method comment
				f.write("\t/**\n")
				if method[4]:
					for l in range(len(method[4])/73+1):
						f.write("\t * %s\n" % method[4][l*73:(l+1)*73])
				if method[2]:
					for param in method[2]:
						f.write("\t * @param %s\n" % param[0]) #param[0]->name
				f.write("\t * @return %s\n\t */\n" % (method[0]=="" or method[0]))
				# if there are no parameter names, something else should appear
				pars = []
				v = ord("a")
				for name, type, value, comment, kind in method[2]:
					#TODO: also use: kind
					if not name:
						name = chr(v)
						v += 1
					if value:
						value = " = " + value
					pars.append((type, '$' + name, value))
				pars = ", ".join([type+" "+name+value for type, name, value in pars])

				vis = method[1] in visibilities and visibilities[method[1]] or ""
				return_type = method[0] == "" or method[0]
				# TODO inheritance type is set to leaf/final by default in Dia. A normal type is need for the default
				#inheritance_type = method[3] == 0 and "abstract " or method[3] == 2 and "final " or ""
				inheritance_type = method[3] == 0 and "abstract " or ""
				static = method[5] and "static " or ""
				f.write("\t%s%s%sfunction %s(%s)" % (static, vis + ' ', inheritance_type, methodname, pars))
				if klass.inheritance_type == "template" or method[3] == 0:
					f.write(";\n\n")
				else:
					f.write("\n\t{\n\t\t// TODO Auto-generated method stub\n\t}\n\n")
			f.write ("}\n\n")
			f.close()
			mainfile.write(" *\t%s\n" % classfile)
		mainfile.write(" */\n")
		mainfile.close()
		ObjRenderer.end_render(self)

	def draw_line(self, *args):pass
	def draw_string(self, *args):pass

# dia-python keeps a reference to the renderer class and uses it on demand
dia.register_export (_("PyDia Code Generation (Python)"), "py", PyRenderer())
dia.register_export (_("PyDia Code Generation (C++)"), "cxx", CxxRenderer())
dia.register_export (_("PyDia Code Generation (Pascal)"), "pas", PascalRenderer())
dia.register_export (_("PyDia Code Generation (Java)"), "java", JavaRenderer())
dia.register_export (_("PyDia Code Generation (JavaScript)"), "js", JsRenderer())
dia.register_export (_("PyDia Code Generation (PHP)"), "php", PhpRenderer())



#    PyDia Self Documentation Series - Part II : Object Types
#    Copyright (c) 2003, Hans Breuer <hans@breuer.org>
#
#        generates a new diagram which contains all the currently
#    registered object types sorted by their containing package
#

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, dia

import gettext
_ = gettext.gettext

def _log(s, append=1) :
	pass
	if append :
		mode = "a"
	else :
		mode = "w"
	f = open("c:\\temp\\otypes.log", mode)
	f.write(s)

def otypes_cb(data, flags) :

	if data:
		diagram = None # we may be running w/o GUI
	else :
		diagram = dia.new("Object Types.dia")
		data = diagram
	layer = data.active_layer

	otypes = dia.registered_types()
	keys = list(otypes.keys())
	keys.sort()

	# property keys w/o overlap
	object_props = ["obj_pos", "obj_bb", "meta"]
	element_props = ["elem_corner", "elem_width", "elem_height"]
	orthconn_props = ["orth_points", "orth_orient", "orth_autoroute"]
	shape_props = ["flip_horizontal", "flip_vertical"]
	# the following are not exclusuve to any objects type
	line_props = ["line_width", "line_style", "line_colour"]
	fill_props = ["fill_colour", "show_background"]
	text_props = ["text_colour", "text_font", "text_height", "text"] # "text_alignment", "text_pos"

	packages = {}
	for s in keys:
		kt = s.split(" - ")
		if len(kt) == 2 :
			if len(kt[0]) == 0 :
				sp = "<unnamed>"
			else :
				sp = kt[0]
			st = kt[1]
		else :
			sp = "<broken>"
			st = kt[0]
		if sp in packages:
			packages[sp].append(st)
		else :
			packages[sp] = [st]

	dtp = dia.get_object_type("UML - LargePackage")
	dtc = dia.get_object_type("UML - Class")
	cy = 0
	maxy = 0
	maxx = 0

	for sp in list(packages.keys()):
		pkg = packages[sp]
		op, h1, h2 = dtp.create(0.0, cy + 1.0)
		op.properties["name"] = sp
		layer.add_object(op)
		cx = 0
		for st in pkg :
			if st == "Group" :
				continue # too special to handle
			oc, h3, h4 = dtc.create(cx + 1.0, cy + 4.0)
			oc.properties["name"] = st
			attrs = []
			# we detect inheritance by common props
			n_object = 0
			n_element = 0
			n_orthconn = 0
			n_shape = 0
			n_line = 0
			n_fill = 0
			n_text = 0
			if st in otypes:
				o_real, h5, h6 = dia.get_object_type(st).create(0,0)
			elif sp + " - " + st in otypes:
				o_real, h5, h6 = dia.get_object_type(sp + " - " + st).create(0,0)
			else :
				o_real = None
				print("Failed to create object", sp, st)
			formal_params = []
			if not o_real is None :
				for p in list(o_real.properties.keys()):
					if p in object_props : n_object = n_object + 1
					elif p in orthconn_props : n_orthconn = n_orthconn + 1
					elif p in element_props : n_element = n_element + 1
					elif p in shape_props : n_shape = n_shape + 1
					elif p in line_props : n_line = n_line + 1
					elif p in text_props : n_text = n_text + 1
					elif p in fill_props : n_fill = n_fill + 1
					else : # don't replicate common props
						attrs.append((p, o_real.properties[p].type, '', '', 0, 0, 0))
				if n_line == len(line_props) :
					formal_params.append(('Line', ''))
				else : # need to add the incomplete set
					for pp in line_props :
						if pp in o_real.properties:
							attrs.append((pp, o_real.properties[pp].type, '', '', 0, 0, 0))
				if n_fill == len(fill_props) :
					formal_params.append(('Fill', ''))
				else :
					for pp in fill_props :
						if pp in o_real.properties:
							attrs.append((pp, o_real.properties[pp].type, '', '', 0, 0, 0))
				if n_text == len(text_props) :
					formal_params.append(('Text', ''))
				else :
					for pp in text_props :
						if pp in o_real.properties:
							attrs.append((pp, o_real.properties[pp].type, '', '', 0, 0, 0))
			if n_orthconn == len(orthconn_props) :
				oc.properties["stereotype"] = "OrthConn"
				oc.properties["fill_colour"] = "light blue"
			elif n_shape == len(shape_props) :
				oc.properties["stereotype"] = "Shape"
				oc.properties["fill_colour"] = "light cyan"
			elif n_element == len(element_props) :
				oc.properties["stereotype"] = "Element"
				oc.properties["fill_colour"] = "light yellow"
			elif n_object == len(object_props) :
				oc.properties["stereotype"] = "Object"
			else :
				print("Huh?", st)
				oc.properties["fill_colour"] = "red"
			oc.properties["attributes"] = attrs
			if len(formal_params) > 0 :
				oc.properties["template"] = 1
				oc.properties["templates"] = formal_params
			layer.add_object(oc)
			# XXX: there really should be a way to safely delete an object. This one will crash:
			# - when the object got added somewhere
			# - any object method gets called afterwards
			if not o_real is None :
				o_real.destroy()
				del o_real
			cx = oc.bounding_box.right
			if maxy < oc.bounding_box.bottom :
				maxy = oc.bounding_box.bottom
			if maxx < cx :
				maxx = cx
			# wrapping too long lines
			if cx > 300 :
				cx = 0
				cy = maxy
		h = op.handles[7]
		# adjust the package size to fit the objects
		op.move_handle(h,(maxx + 1.0, maxy + 1.0), 0, 0)
		cy = maxy + 2.0
		maxx = 0 # every package a new size
	data.update_extents()
	if diagram :
		diagram.display()
		diagram.flush()
	# make it work standalone
	return data

dia.register_action ("HelpOtypes", _("Dia Object _Types"),
                     "/ToolboxMenu/Help/HelpExtensionStart",
                     otypes_cb)



import sys, dia

import gettext
_ = gettext.gettext

# sys.path.insert(0, 'd:/graph/dia/dia')

##
# \brief A simple example renderer implemented in Python
# \extends _DiaPyRenderer
# \ingroup PyDia
class DumpRenderer :
	## \brief Constructor
	def __init__ (self) :
		pass
	## \brief Start rendering
	# For non-interactive renderers it is guaranteed that this function is
	# called before every drawing function. It should be used to clear the
	# state from previous use, because the renderer class gets only created
	# once per Dia session
	def begin_render (self, data, filename) :
		# DiagramData
		self.f = open(filename, "w")
		self.f.write("Data: " + str(data) + "\n")
		self.f.write("Extents: "+ str(data.extents) + "\n")
		self.f.write("active_layer: " + str(data.active_layer) + " " \
				+ data.active_layer.name + "\n")
		#self.f.write("grid .width: " + str(data.grid.width) \
		#			+ " .height" + str(data.grid.height) \
		#			+ "visible: " + str(data.visible) + "\n")
	## \brief End rendering
	# For non-interactive renderers it is guaranteed that all drawing
	# is finished with this call
	def end_render (self) :
		self.f.close()
	## \brief Remember the line width
	def set_linewidth (self, width) :
		self.line_width = width
	## \brief Remember the line caps
	def set_linecaps (self, mode) :
		self.line_caps = mode
	## \brief Remember the line join
	def set_linejoin (self, mode) :
		self.line_join = mode
	## \brief Remember the line style and dash length
	def set_linestyle (self, style, length) :
		self.line_style = style
		self.dash_length = length
	## \brief Remember the fill style
	def set_fillstyle (self, style) :
		self.fill_style = style
	## \brief Remember the font
	def set_font (self, font, size) :
		self.font = font
	## \brief Draw a straight line from start to end with color
	# @param start A point in diagram coordinate system
	# @param end A point in diagram coordinate system
	# @param color The color to use for the line
	# The addtional paramters needed for the drawing are given
	# by one of the set_*() function above
	def draw_line (self, start, end, color) :
		self.f.write("draw_line:" + str(start) + str(end) + str(color) + "\n")
	## \brief Draw a polyline
	# @param points An array of points in the diagram coordinate system
	# @param color The color to use for the line
	def draw_polyline (self, points, color) :
		self.f.write("draw_polyline: " + str(color) + "\n")
		for pt in points :
			self.f.write ("\t" + str(pt) + "\n")
	## \brief Draw a polygon
	# @param points An array of points in the diagram coordinate system
	# @param fill The color to use for the interior
	# @param stroke The color to use for the line
	def draw_polygon (self, points, fill, stroke) :
		self.f.write("draw_polygon: " + str(fill) + str(stroke) + "\n")
		for pt in points :
			self.f.write ("\t" + str(pt) + "\n")
	## \brief Draw a rectangle
	def draw_rect (self, rect, color) :
		self.f.write("draw_rect: " + str(rect) + str(fill) + str(stroke) + "\n")
	## \brief Draw an arc
	def draw_arc (self, center, width, height, angle1, angle2, color) :
		self.f.write("draw_arc: " + str(center) + ";" \
				+ str(width) + "x" + str(height) + ";" \
				+ str(angle1) + "," + str(angle2) + ";" + str(color) + "\n")
	## \brief Fill an arc
	def fill_arc (self, center, width, height, angle1, angle2, color) :
		self.f.write("fill_arc: " + str(center) + ";" \
				+ str(width) + "x" + str(height) + ";" \
				+ str(angle1) + "," + str(angle2) + ";" + str(color) + "\n")
	## \brief Draw an ellipse
	def draw_ellipse (self, center, width, height, fill, stroke) :
		self.f.write("draw_ellipse: " + str(center) \
				+ str(width) + "x" +str(height)
				+ ";" + str(fill) + ";" + str(stroke) + "\n")
	## \brief Draw a bezier line
	def draw_bezier (self, bezpoints, color) :
		self.f.write("draw_bezier: " + str(color) + "\n")
		for pt in bezpoints :
			self.f.write ("\t" + str(pt) + "\n")
	## \brief Fill a bezier shape
	def draw_beziergon (self, bezpoints, fill, stroke) :
		self.f.write("draw_bezier: " + str(fill) + "; " + str(stroke) + "\n")
		for pt in bezpoints :
			self.f.write ("\t" + str(pt) + "\n")
	## \brief Draw a string
	def draw_string (self, text, pos, alignment, color) :
		self.f.write("draw_string: [" + text + "]; " + str(pos) \
				+ str(alignment) + "; " +str(color))
	## \brief Draw an image
	def draw_image (self, point, width, height, image) :
		self.f.write("draw_image: " + str(point) + str(width) + "x" +str(height) \
				+ " " + image.filename + "\n")
		self.f.write("<rgb_data>" + image.rgb_data + "</rgb_data>")
		self.f.write("<mask_data>" + image.mask_data + "</mask_data>")

## \brief Register the renderer with Dia's export system
# dia-python keeps a reference to the renderer class and uses it on demand
dia.register_export (_("PyDia Render Export"), "diapyr", DumpRenderer())
